#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''

from ._anaerobic_baffled_reactor import *
from ._anaerobic_digestion import *
from ._biogas_combustion import *
from ._bst_units import *
from ._clarifier import *
from ._component_splitter import *
from ._crop_application import *
from ._decay import *
from ._drying_bed import *
from ._electrochemical_cell import *
from ._excretion import *
from ._lagoon import *
from ._liquid_treatment_bed import *
from ._lumped_cost import *
from ._pit_latrine import *
from ._sedimentation_tank import *
from ._sludge_separator import *
from ._suspended_growth_bioreactor import *
from ._toilet import *
from ._trucking import *
from ._uddt import *


from . import (
    _anaerobic_baffled_reactor,
    _anaerobic_digestion,
    _biogas_combustion,
    _bst_units,
    _clarifier,
    _component_splitter,
    _crop_application,
    _decay,
    _drying_bed,
    _electrochemical_cell,
    _excretion,
    _lagoon,
    _liquid_treatment_bed,
    _lumped_cost,
    _pit_latrine,
    _sedimentation_tank,
    _sludge_separator,
    _suspended_growth_bioreactor,
    _toilet,
    _trucking,
    _uddt,
    )


__all__ = (
    *_anaerobic_baffled_reactor.__all__,
    *_anaerobic_digestion.__all__,
    *_biogas_combustion.__all__,
    *_bst_units.__all__,
    *_clarifier.__all__,
    *_component_splitter.__all__,
    *_crop_application.__all__,
    *_decay.__all__,
    *_drying_bed.__all__,
    *_electrochemical_cell.__all__,
    *_excretion.__all__,
    *_lagoon.__all__,
    *_liquid_treatment_bed.__all__,
    *_lumped_cost.__all__,
    *_pit_latrine.__all__,
    *_sedimentation_tank.__all__,
    *_sludge_separator.__all__,
    *_suspended_growth_bioreactor.__all__,
    *_toilet.__all__,
    *_trucking.__all__,
    *_uddt.__all__,
           )



# -*- coding: utf-8 -*-
'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Joy Cheung <joycheung1994@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.


Reference
---------
[1] Takács, I.; Patry, G. G.; Nolasco, D. A Dynamic Model of the 
Clarification-Thickening Process. Water Res. 1991, 25 (10), 1263–1271. 
https://doi.org/10.1016/0043-1354(91)90066-Y.

'''

__all__ = ('FlatBottomCircularClarifier', )

from .. import SanUnit
from math import exp
# from sympy import symbols, lambdify, Matrix
from scipy.integrate import solve_ivp
from warnings import warn
import numpy as np

def _settling_flux(X, v_max, v_max_practical, X_min, rh, rp):
    X_star = max(X-X_min, 0)
    v = min(v_max_practical, v_max*(exp(-rh*X_star) - exp(-rp*X_star)))
    return X*max(v, 0)

class FlatBottomCircularClarifier(SanUnit):
    
    _N_ins = 1
    _N_outs = 2
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, 
                 init_with='WasteStream', sludge_flow_rate=2000, 
                 surface_area=1500, height=4, N_layer=10, feed_layer=4, 
                 X_threshold=3000, v_max=474, v_max_practical=250, 
                 rh=5.76e-4, rp=2.86e-3, fns=2.28e-3, **kwargs):
        SanUnit.__init__(self, ID, ins, outs, thermo, init_with)
        self._Qs = sludge_flow_rate
        self._V = surface_area * height
        self._A = surface_area
        self._hj = height/N_layer
        self._N_layer = N_layer
        self._feed_layer = feed_layer
        self._v_max = v_max
        self._v_max_p = v_max_practical
        self._X_t = X_threshold
        self._rh = rh
        self._rp = rp
        self._fns = fns
        for attr, value in kwargs.items():
            setattr(self, attr, value)
        self._init_X = None   #TSS
        self._init_Z = None   #TDS
        
    def _run(self, t_bound=10, steady_state=False, cache_state=True):
        inf = self.ins[0]
        X_in = inf.get_TSS()
        Q_in = inf.get_total_flow('m3/d')
        eff, sludge = self.outs
        eff.copy_like(inf)
        sludge.copy_like(inf)
        Q_s = self._Qs
        Q_e = Q_in - Q_s
        n = self._N_layer
        jf = self._feed_layer - 1
        if jf not in range(self._N_layer): 
            raise ValueError(f'feed layer {self._feed_layer} is out of range.'
                             f'must be an integer between 1 and {self._N_layer}.')
        if self._init_X is not None: X_0 = self._init_X
        else: X_0 = np.ones(n)*X_in  
        X_min = X_in * self._fns
        def dX_dt(t, X):
            flow_out = X*Q_e*[j <= jf for j in range(n)] + X*Q_s*[j >= jf for j in range(n)]
            flow_in = np.array([flow_out[j+1] if j < jf else flow_out[j-1] for j in range(n)])
            flow_in[jf] = Q_in * X_in            
            VX = [_settling_flux(x, self._v_max, self._v_max_p, X_min, self._rh, self._rp) for x in X]
            J = [VX[j] if X[j+1] <= self._X_t and j < jf else min(VX[j], VX[j+1]) for j in range(n-1)]
            settle_out = np.array(J + [0])
            settle_in = np.array([0] + J)
            dXdt = ((flow_in - flow_out)/self._A + settle_in - settle_out)/self._hj
            return dXdt
        def limit(t, X):
            dXdt = dX_dt(0, X)
            if np.allclose(dXdt, np.zeros(len(X)), atol=1e-2): return 0
            else: return 1
        limit.terminal = True
        
        sol = solve_ivp(dX_dt, (0, t_bound), X_0, events=limit)
        X_out = sol.y.transpose()[-1]
        if steady_state:
            while len(sol.t_events) == 0:
                sol = solve_ivp(dX_dt, (0, t_bound), X_out, events=limit)
                X_out = sol.y.transpose()[-1]
        else:
            if len(sol.t_events) == 0: 
                warn(f'{self.ID} did not reach steady state in this run.')
        if cache_state: self._init_X = X_out
        
        cmps = self.components
        TSS_ratio_eff_in = X_out[1]*Q_e/(X_in*Q_in)
        eff_mass_flow = (TSS_ratio_eff_in*cmps.x + Q_e/Q_in*(1-cmps.x))*inf.mass
        eff.set_flow(eff_mass_flow, 'kg/hr', cmps.IDs)
        sludge.separate_out(eff)
    
    def _design(self):
        pass


# -*- coding: utf-8 -*-
'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Joy Cheung <joycheung1994@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''

from .. import SanUnit, WasteStream, Process, Processes
from ._clarifier import _settling_flux
from sympy import symbols, lambdify, Matrix
from scipy.integrate import solve_ivp
from warnings import warn
from math import floor, ceil
import numpy as np
import pandas as pd

__all__ = ('CSTR', 
           'SBR', 
           # 'PFR',
           )

def _add_aeration_to_growth_model(aer, model):
    if isinstance(aer, Process):
        processes = Processes(model.tuple)
        processes.append(aer)
        processes.compile()
        processes.set_parameters(**aer.parameters, **model.parameters)
    else:
        processes = model
        processes.compile()
    return processes

#%%
class CSTR(SanUnit):
    '''
    A single continuous stirred tank reactor.

    Parameters
    ----------
    ID : str
        ID for the reactor.
    ins : :class:`WasteStream`
        Influents to the reactor. Can be an array of up to 3 WasteStream objects by
        default, typically wastewater to be treated, recycled effluent, recycled 
        activated sludge.
    outs : :class:`WasteStream`
        Treated effluent.
    V_max : float
        Designed volume, in [m^3]. The default is 1000.
    aeration : float or :class:`Process`, optional
        Aeration setting. Either specify a targeted dissolved oxygen concentration
        in [mg O2/L] or provide a :class:`Process` object to represent aeration,
        or None for no aeration. The default is 2.0.
    DO_ID : str, optional
        The :class:`Component` ID for dissolved oxygen, only relevant when the 
        reactor is aerated. The default is 'S_O2'.
    suspended_growth_model : :class:`Processes`, optional
        The suspended growth biokinetic model. The default is None.

    '''

    _N_ins = 3
    _N_outs = 1
    _ins_size_is_fixed = False
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, init_with='WasteStream', 
                 V_max=1000, aeration=2.0, DO_ID='S_O2',
                 suspended_growth_model=None, **kwargs):
        SanUnit.__init__(self, ID, ins, outs, thermo, init_with)
        self._V_max = V_max
        self._aeration = aeration
        self._DO_ID = DO_ID
        self._model = suspended_growth_model
        for attr, value in kwargs.items():
            setattr(self, attr, value)
        self._init_C = None
        
    def _run(self, t_bound=10, steady_state=False, cache_state=True):
        isa = isinstance        
        if self._model is not None:
            mixed = WasteStream()
            mixed.mix_from(self.ins)
            treated = self.outs[0]
            treated.copy_like(mixed)
            Q = mixed.get_total_flow('m3/d')
            tau = self._V_max / Q
            C_in = mixed.mass / mixed.F_vol * 1e3    # concentrations in g/m3
            if self._init_C is not None: C_0 = self._init_C
            else: C_0 = C_in            
            C = list(symbols(self.components.IDs))
            
            processes = _add_aeration_to_growth_model(self._aeration, self._model)                               
            mass_balance_terms = list(zip(C_in, C, processes.production_rates.rate_of_production))
            C_dot_eqs = [(cin-c)/tau + r for cin, c, r in mass_balance_terms]
            if isa(self._aeration, (float, int)):
                i = self.components.index(self._DO_ID)
                C_0[i] = C_in[i] = self._aeration
                C_dot_eqs[i] = 0
            
            def dC_dt(t, y):
                C_dot = lambdify(C, C_dot_eqs)
                return C_dot(*y)
            def limit(t, y):
                dCdt = np.array(dC_dt(0, y))
                if np.allclose(dCdt, np.zeros(len(y)), atol=1e-3): return 0
                else: return 1
            limit.terminal = True            
            J = Matrix(dC_dt(None, C)).jacobian(C)
            def J_func(t, y):
                J_func = lambdify(C, J)
                return J_func(*y)            
            
            sol = solve_ivp(dC_dt, (0, t_bound), C_0, method='BDF', jac=J_func, events=limit)
            C_out = np.array([max(y, 0) for y in sol.y.transpose()[-1]])
            if steady_state:
                while len(sol.t_events) == 0:
                    sol = solve_ivp(dC_dt, (0, t_bound), C_out, method='BDF', jac=J_func, events=limit)
                    C_out = np.array([max(y, 0) for y in sol.y.transpose()[-1]])
            else:
                if len(sol.t_events) == 0: 
                    warn(f'{self.ID} did not reach steady state in this run.')
            if cache_state: self._init_C = C_out
            treated.set_flow(C_out*treated.F_vol, 'g/hr', self.components.IDs)
        else:
            raise RuntimeError(f'{self.ID} was initiated without a suspended growth model.')
    
    def _design(self):
        pass
    
    @property
    def init_state(self):
        return dict(zip(self.components.IDs, self._init_C))
    
    @init_state.setter
    def init_state(self, C):
        if len(C) == len(self.components.IDs):
            self._init_C = np.asarray(C)
        else: 
            raise ValueError(f'Must be a 1D array of length {len(self.components.IDs)}')

class SBR(SanUnit):
    '''
    Sequential batch reactors operated in parallel. The number of reactors is 
    determined by operation cycle and influent flowrate. 

    Parameters
    ----------
    ID : str
        ID for the reactors. The default is ''.
    ins : :class:`WasteStream`
        Influent to the reactor. Expected number of influent is 1.
    outs : :class:`WasteStream`
        Treated effluent and wasted sludge.
    surface_area : float, optional
        Surface area of the reactor bottom, in [m^2]. The reactor is assumed 
        to be cylinder. The default is 1500.
    height : float, optional
        Height of the reactor, in [m]. The default is 4.
    operation_cycle : iterable of float, optional
        Operation cycle of the SBR, time for each stage specified in [h]. There
        are 7 stages: 1 - fill, 2 - fill, 3 - mix, 4 - mix, 5 - settle, 6 - decant, 
        7 - desludge. The first 4 stages are modeled as a biological reactor. 
        The 5th stage is modeled as a 1D N-layer settler. The last 2 stages are 
        assumed inactive. The default is (0.5, 1.5, 2.0, 0, 1.0, 0.5, 0.1).
    aeration : iterable of float and/or :class:`Process`, optional
        Aeration settings for the first 4 stages of the operation cycle. Either 
        specify a targeted dissolved oxygen concentration in [mg O2/L] or provide 
        a :class:`Process` object to represent aeration, or None for no aeration. 
        The default is (None, None, None, 2.0).
    DO_ID : str, optional
        The :class:`Component` ID for dissolved oxygen, only relevant when the 
        reactor is aerated. The default is 'S_O2'.
    suspended_growth_model : :class:`Processes`, optional
        The suspended growth biokinetic model. The default is None.
    N_layer : int, optional
        The number of layers to model settling. The default is 10.
    pumped_flow : float, optional
        Designed effluent flowrate, in [m^3/d]. The default is None.
    underflow : float, optional
        Designed wasted activated sludge flowrate, in [m^3/d]. The default is None.
    X_threshold : float, optional
        Threshold suspended solid cocentration, in [g/m^3]. The default is 3000.
    v_max : float, optional
        Maximum theoretical (i.e. Vesilind) settling velocity, in [m/d]. The 
        default is 474.
    v_max_practical : float, optional
        Maximum practical settling velocity, in [m/d]. The default is 250.
    rh : float, optional
        Hindered zone settling parameter in the double-exponential settling velocity
        function, in [m^3/g]. The default is 5.76e-4.
    rp : float, optional
        Flocculant zone settling parameter in the double-exponential settling velocity
        function, in [m^3/g]. The default is 2.86e-3.
    fns : float, optional
        Non-settleable fraction of the suspended solids, dimensionless. Must be within 
        [0, 1], The default is 2.28e-3.

    References
    ----------
    .. [1] Takács, I.; Patry, G. G.; Nolasco, D. A Dynamic Model of the Clarification
        -Thickening Process. Water Res. 1991, 25 (10), 1263–1271. 
        https://doi.org/10.1016/0043-1354(91)90066-Y.

    '''
        
    _N_ins = 1
    _N_outs = 2
    
    def __init__(self, ID='', ins=None, outs=(), surface_area=1500, height=4, 
                 operation_cycle=(0.5, 1.5, 2.0, 0, 1.0, 0.5, 0.1), 
                 aeration=(None, None, None, 2.0), DO_ID='S_O2',
                 suspended_growth_model=None, N_layer=10, 
                 pumped_flow=None, underflow=None,
                 X_threshold=3000, v_max=474, v_max_practical=250, 
                 rh=5.76e-4, rp=2.86e-3, fns=2.28e-3, **kwargs):
        SanUnit.__init__(self, ID, ins, outs)
        self._V = surface_area * height
        self._A = surface_area
        self._h = height
        self._operation_cycle = operation_cycle
        self._aeration = aeration
        self._DO_ID = DO_ID
        self._model = suspended_growth_model
        self._N_layer = N_layer
        self._Q_e = pumped_flow
        self._Q_WAS = underflow
        self._X_t = X_threshold
        self._v_max = v_max
        self._v_max_p = v_max_practical
        self._rh = rh
        self._rp = rp
        self._fns = fns
        for attr, value in kwargs.items():
            setattr(self, attr, value)
        self._init_Vas = None
        self._init_Cas = None
        self._dynamic_composition = None
        
    @property
    def operation_cycle(self):
        return dict(zip(('fill_1', 'fill_2', 'mix_1', 'mix_2', 'settle', 'decant', 'desludge'),
                        self._operation_cycle))
    @property
    def total_cycle_time(self):
        return sum(self._operation_cycle)
    
    @property
    def aeration(self):
        return dict(zip(('fill_1', 'fill_2', 'mix_1', 'mix_2'),
                        self._aeration[:4]))
    
    @property
    def C_t(self):
        if self._dynamic_composition:
            return pd.DataFrame(self._dynamic_composition, 
                                columns = ['Time[d]'] + list(self.components.IDs))
        else: return None
    
    def _run(self, cache_state=True):
        isa = isinstance        
        if self._model is not None:
            inf = self.ins[0]
            Q_in = inf.get_total_flow('m3/d')
            eff, sludge = self.outs
            eff.copy_like(inf)
            sludge.copy_like(inf)
            C_in = inf.mass / inf.F_vol * 1e3    # concentrations in g/m3
            cmps = self.components
            C = list(symbols(cmps.IDs))
            if self._init_Vas is not None:
                V_0 = self._init_Vas
                C_0 = self._init_Cas
            else: 
                V_0 = 0
                C_0 = C_in
            n = self._N_layer
            if self._aeration.count(None) == len(self._aeration): 
                Vmax = self._V
                hj = self._h/n
            else: 
                Vmax = self._V*0.75
                hj = self._h*0.75/n
            
            # fill and mix/aerate stages
            T_fill = (Vmax - V_0)/Q_in # maximum total fill time in day
            T = [t/24 for t in self._operation_cycle]  # operation cycle in day
            if T_fill <= T[0]: 
                schedule = [T_fill, T[0]-T_fill] + T[1:4]
                aer = [self._aeration[0], self._aeration[0]] + list(self._aeration[1:4])
                fill = [True] + [False]*4
                V_total = Vmax
            elif T_fill <= T[0]+T[1]:
                schedule = [T[0], T_fill-T[0], T[0]+T[1]-T_fill] + T[2:4]
                aer = list(self._aeration[:2]) + [self._aeration[1]] + list(self._aeration[2:4])
                fill = [True]*2 + [False]*3
                V_total = Vmax
            else:
                schedule = T[:4]
                aer = list(self._aeration[:4])
                fill = [True]*2 + [False]*2
                V_total = Q_in*(T[0]+T[1])+V_0
                hj = V_total/self._V*self._h/n
            
            for i in range(1, len(schedule)):
                if fill[-i] == fill[-i-1] and aer[-i] == aer[-i-1]:
                    schedule[-i-1] += schedule[-i]
                    schedule[-i] = 0
            
            t_arr = np.array([])
            y_mat = np.ndarray([])
            for i in range(len(schedule)):
                if schedule[i] > 0:
                    dC_dt, J_func = self._compile_dC_dt(V_0, Q_in, C_in, C, fill[i], aer[i])
                    if isa(aer[i], (float, int)): C_0[cmps.index(self._DO_ID)] = aer[i]
                    sol = solve_ivp(dC_dt, (0, schedule[i]), C_0, method='BDF', jac=J_func)
                    C_0 = sol.y.transpose()[-1]
                    V_0 += Q_in * schedule[i] * fill[i]
                    t_arr = np.concatenate((t_arr, sol.t + t_arr[-1]))
                    y_mat = np.hstack((y_mat, sol.y))
            self._dynamic_composition = np.vstack((t_arr, y_mat)).transpose()
            
            # settle, decant, desludge
            eff.set_flow(C_0*eff.F_vol, 'g/hr', self.components.IDs)
            X_0 = eff.get_TSS()
            X_min = X_0 * self._fns
            T_settle = T[4]
            def dX_dt(t, X):         
                VX = [_settling_flux(x, self._v_max, self._v_max_p, X_min, self._rh, self._rp) for x in X]
                J = [VX[j] if X[j+1] <= self._X_t else min(VX[j], VX[j+1]) for j in range(n-1)]
                settle_out = np.array(J + [0])
                settle_in = np.array([0] + J)
                dXdt = (settle_in - settle_out)/hj
                return dXdt
            sol = solve_ivp(dX_dt, (0, T_settle), np.ones(n)*X_0)
            X = sol.y.transpose()[-1]
            
            V_eff = min(T[5]*self._Q_e, V_total*(n-1)/n)
            n_eff = V_eff/V_total
            w_eff = np.array([1]*floor(n_eff)+[n_eff-floor(n_eff)])
            X_eff = np.average(X[:ceil(n_eff)], weights=w_eff)
            eff_mass_flow = (X_eff/X_0*cmps.x + (1-cmps.x))*C_0*V_eff/T[5]
            eff.set_flow(eff_mass_flow, 'g/d', cmps.IDs)
            
            V_was = min(T[6]*self._Q_WAS, V_total-V_eff)
            X_as = (V_total*X_0 - V_eff*X_eff) / (V_total-V_eff)
            C_as = (X_as/X_0*cmps.x + (1-cmps.x))*C_0
            was_mass_flow = C_as*V_was/T[6]
            sludge.set_flow(was_mass_flow, 'g/d', cmps.IDs)
            
            if cache_state: 
                self._init_Vas = V_total - V_eff - V_was
                self._init_Cas = C_as
        else: 
            raise RuntimeError(f'{self.ID} was initiated without a suspended growth model.')

    def _design(self):
        pass
    
    def _compile_dC_dt(self, V0, Qin, Cin, C, fill, aer):
        isa = isinstance
        processes = _add_aeration_to_growth_model(aer, self._model)
        if fill:
            t = symbols('t')
            mass_balance_terms = list(zip(Cin, C, processes.production_rates.rate_of_production))
            C_dot_eqs = [(cin-c)/(t+V0/Qin) + r for cin, c, r in mass_balance_terms]
            if isa(aer, (float, int)): C_dot_eqs[self.components.index(self._DO_ID)] = 0
            def dC_dt(t, y):
                C_dot = lambdify([t]+C, C_dot_eqs)
                return C_dot(t, *y)
            J = Matrix(dC_dt(t, C)).jacobian(C)
        else:
            C_dot_eqs = processes.production_rates.rate_of_production
            if isa(aer, (float, int)): C_dot_eqs[self.components.index(self._DO_ID)] = 0
            def dC_dt(t, y):
                C_dot = lambdify(C, C_dot_eqs)
                return C_dot(*y)
            J = Matrix(dC_dt(None, C)).jacobian(C)
        def J_func(t, y):
            J_func = lambdify(C, J)
            return J_func(*y)
        return (dC_dt, J_func)

# class PFR(SanUnit):
    
#     _N_ins = 1
#     _N_outs = 2
    
#     def __init__(self, ID='', ins=None, outs=(), **kwargs):
#         SanUnit.__init__(self, ID, ins, outs)
    
#     def _run(self, steady_state=True):
#         pass
    
#     def _design(self):
#         pass
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''


# %%

import sys
import pandas as pd
from warnings import warn
from thermosteam.utils import registered
from . import currency, SanStream, WasteStream, ImpactIndicator
from .utils import (
    auom, parse_unit, copy_attr,
    format_number as f_num
    )


__all__ = ('ImpactItem', 'StreamImpactItem')

def check_source(item, return_item=False, raise_error=True):
    if item.source:
        if raise_error:
            raise ValueError(f'This impact item is copied from {item.source.ID}, '
                             'value cannot be set.')
        else:
            item = item.source

    if return_item:
        return item

@registered(ticket_name='item')
class ImpactItem:
    '''
    A class for calculation of environmental impacts.

    Parameters
    ----------
    ID : str
        ID of the impact item.
    functional_unit : str
        Functional unit of the impact item.
    price : float
        Price of the item per functional unit.
    price_unit : str
        Unit of the price.
    source : :class:`ImpactItem`
        If provided, all attributions and properties of this impact item will
        be copied from the provided source.
    indicator_CFs : kwargs
        Impact indicators and their characteriziation factors (CFs),
        can be in the form of str=float or str=(float, unit).

    Tip
    ---
    :class:`ImpactItem` should be used for environmental impacts associated with
    construction and transportation.
    For impacts associated with streams (e.g., chemicals, wastes, emissions),
    use :class:`StreamImpactItem` instead.


    Examples
    --------
    Firstly make impact indicators.

    >>> import qsdsan as qs
    >>> GWP = qs.ImpactIndicator('GlobalWarming', alias='GWP', unit='kg CO2-eq')
    >>> FEC = qs.ImpactIndicator('FossilEnergyConsumption', alias='FEC', unit='MJ')

    We can make impact items in different ways (numbers are made up).

    >>> Steel = qs.ImpactItem('Steel', 'kg', GWP=2.55)
    >>> Steel.show()
    ImpactItem      : Steel [per kg]
    Price           : None USD
    ImpactIndicators:
                               Characterization factors
    GlobalWarming (kg CO2-eq)                      2.55
    >>> # Unit will be automatically converted to match the unit of the impact indicator
    >>> Electricity = qs.ImpactItem('Electricity', functional_unit='kWh',
    ...                             GWP=(480, 'g CO2-eq'), FEC=(5926, 'kJ'))
    >>> Electricity.show()
    ImpactItem      : Electricity [per kWh]
    Price           : None USD
    ImpactIndicators:
                                  Characterization factors
    GlobalWarming (kg CO2-eq)                         0.48
    FossilEnergyConsumption (MJ)                      5.93
    >>> # Note that 5.93 appearing instead of 5.926 is for nicer print
    >>> Electricity.CFs
    {'GlobalWarming': 0.48, 'FossilEnergyConsumption': 5.926}
    >>> # Get all impact items
    >>> qs.ImpactItem.get_all_items()
    {'Steel': <ImpactItem: Steel>, 'Electricity': <ImpactItem: Electricity>}

    You can make copies of impact items and choose to link to the source or not.

    >>> Steel2 = Steel.copy('Steel2', set_as_source=True)
    >>> Steel2.CFs['GlobalWarming']
    2.55
    >>> Steel3 = Steel.copy('Steel3', set_as_source=False)
    >>> Steel3.CFs['GlobalWarming']
    2.55
    >>> # Once linked, CFs of the copy will update with the source
    >>> Steel.CFs['GlobalWarming'] = 2
    >>> Steel.CFs['GlobalWarming']
    2
    >>> Steel2.CFs['GlobalWarming']
    2
    >>> Steel3.CFs['GlobalWarming']
    2.55
    >>> # Update the copy won't update the source
    >>> Steel2.CFs['GlobalWarming'] = 5
    >>> Steel.CFs['GlobalWarming']
    2
    >>> Steel2.CFs['GlobalWarming']
    2

    Manage the registry.

    >>> qs.ImpactItem.get_all_items()
    {'Steel': <ImpactItem: Steel>,
     'Electricity': <ImpactItem: Electricity>,
     'Steel2': <ImpactItem: Steel2>,
     'Steel3': <ImpactItem: Steel3>}
    >>> Steel2.deregister()
    The impact item "Steel2" has been removed from the registry.
    >>> qs.ImpactItem.get_all_items()
    {'Steel': <ImpactItem: Steel>,
     'Electricity': <ImpactItem: Electricity>,
     'Steel3': <ImpactItem: Steel3>}
    >>> Steel2.register()
    The impact item "Steel2" has been added to the registry.
    >>> qs.ImpactItem.get_all_items()
    {'Steel': <ImpactItem: Steel>,
     'Electricity': <ImpactItem: Electricity>,
     'Steel3': <ImpactItem: Steel3>,
     'Steel2': <ImpactItem: Steel2>}
    >>> qs.ImpactItem.clear_registry()
    All impact items have been removed from registry.
    >>> qs.ImpactItem.get_all_items()
    {}
    '''

    _items = {}

    __slots__ = ('_ID', '_functional_unit', '_price', '_CFs', '_source')

    def __init__(self, ID='', functional_unit='kg', price=0., price_unit='',
                 source=None, **indicator_CFs):

        self._register(ID)

        if source:
            self.source = source
        else:
            self._source = None
            self._functional_unit = auom(functional_unit)
            self._update_price(price, price_unit)
            self._CFs = {}
            for indicator, value in indicator_CFs.items():
                try:
                    CF_value, CF_unit = value # unit provided for CF
                    self.add_indicator(indicator, CF_value, CF_unit)
                # except: breakpoint()
                except Exception as e:
                    if 'unpack' in str(sys.exc_info()[1]):
                        self.add_indicator(indicator, value)
                    else:
                        raise e


    # This makes sure it won't be shown as memory location of the object
    def __repr__(self):
        return f'<ImpactItem: {self.ID}>'

    def show(self):
        '''Show basic information of this impact item.'''
        info = f'ImpactItem      : {self.ID} [per {self.functional_unit}]'
        if self.source:
            info += f'\nSource          : {self.source.ID}'
        info += f'\nPrice           : {f_num(self.price)} {currency}'
        info += '\nImpactIndicators:'
        print(info)
        if len(self.CFs) == 0:
            print(' None')
        else:
            index = pd.Index((i.ID+' ('+i.unit+')' for i in self.indicators))
            df = pd.DataFrame({
                'Characterization factors': tuple(self.CFs.values())
                },
                index=index)
            print(df.to_string())

    _ipython_display_ = show

    def _update_price(self, price=0., unit=''):
        source_item = check_source(self, True)
        if not unit or unit == currency:
            source_item._price = float(price)
        else:
            converted = auom(unit).convert(float(price), currency)
            source_item._price = converted

    def add_indicator(self, indicator, CF_value, CF_unit=''):
        '''Add an indicator with charactorization factor values.'''
        source_item = check_source(self, True)
        if isinstance(indicator, str):
            indicator = ImpactIndicator.get_indicator(indicator)

        CF_unit2 = CF_unit.replace(' eq', '-eq')

        if CF_unit and CF_unit != indicator.unit and CF_unit2 != indicator.unit:
            try:
                CF_value = auom(parse_unit(CF_unit)[0]). \
                    convert(CF_value, indicator._ureg_unit.units)
            except:
                raise ValueError(f'Conversion of the given unit {CF_unit} to '
                                  f'the defaut unit {indicator.unit} is not supported.')

        source_item._CFs[indicator.ID] = CF_value

    def remove_indicator(self, indicator):
        '''
        Remove an indicator from this impact item.

        Parameters
        ----------
        indicator : str or :class:`~.ImpactIndicator`
            :class:`~.ImpactIndicator` or its ID.
        '''
        ID = indicator if isinstance(indicator, str) else indicator.ID
        check_source(self, True).CFs.pop(ID)
        print(f'The impact indicator "{ID}" has been removed.')


    def copy(self, new_ID='', set_as_source=False):
        '''
        Return a new :class:`ImpactItem` object with the same settings.

        Parameters
        ----------
        new_ID : str
            ID of the new impact item.
        set_as_source : bool
            Whether to set the original impact item as the source.
        '''

        new = ImpactItem.__new__(ImpactItem)
        new.__init__(new_ID)

        if set_as_source:
            if self.source:
                new.source = self.source
            else:
                new.source = self
        else:
            new = copy_attr(new, self, skip=('_ID', '_source'))

        return new

    __copy__ = copy

    def register(self, print_msg=True):
        '''Add this impact item to the registry.'''
        self.registry.register_safely(self.ID, self)
        if print_msg:
            print(f'The impact item "{self.ID}" has been added to the registry.')

    def deregister(self, print_msg=True):
        '''Remove this impact item from the registry.'''
        self.registry.discard(self.ID)
        if print_msg:
            print(f'The impact item "{self.ID}" has been removed from the registry.')

    @classmethod
    def clear_registry(cls, print_msg=True):
        '''Remove all existing impact items from the registry.'''
        cls.registry.clear()
        if print_msg:
            print('All impact items have been removed from registry.')


    @classmethod
    def get_all_items(cls):
        '''Get all impact items.'''
        return cls.registry.data

    @classmethod
    def get_item(cls, ID):
        '''Get an item by its ID.'''
        return cls.get_all_items().get(ID)

    @classmethod
    def _load_from_df(cls, name, df):
        if name.lower() == 'info':
            for num in df.index:
                new = cls.__new__(cls)
                new.__init__(ID=df.iloc[num].ID,
                             functional_unit=df.iloc[num].functional_unit)
        else:
            for num in df.index:
                item = cls.get_item(df.iloc[num].ID)
                item.add_indicator(indicator=name,
                                   CF_value=float(df.iloc[num].expected),
                                   CF_unit=df.iloc[num].unit)

    @classmethod
    def load_items_from_excel(cls, path_or_dict, index_col=None):
        '''Same as :func:`load_from_file`, has been deprecated.'''
        warn('`load_items_from_excel` has been deprecated, '
             'please use `load_from_file` instead.', stacklevel=2)
        cls.load_from_file(path_or_dict, index_col)


    @classmethod
    def load_from_file(cls, path_or_dict, index_col=None):
        '''
        Load impact items from an Excel file or a :class:`pandas.DataFrame`.

        This Excel should have multiple sheets:

            - The "info" sheet should have two columns: "ID" (e.g., Cement) \
            and "functional_unit" (e.g., kg) of different impact items.

            - The remaining sheets should contain characterization factors of \
            impact indicators.

                - Name of the sheet should be the ID (e.g., GlobalWarming) or \
                alias (e.g., GWP) of the indicator.

                - Each sheet should have at least two columns: "unit" (e.g., kg CO2-eq) \
                and "expected" (values) of the CF.

                - You can also have additional columns to be used for other purpose \
                (e.g., uncertainty analysis).

        .. note::

            This function is just one way to batch-load impact items,
            you can always write your own function that fits your datasheet format,
            as long as it provides all the information to construct new impact items.


        Parameters
        ----------
        path_or_dict : str or dict of :class:`pandas.DataFrame`
            A dict of DataFrame or complete path of the datasheet in xls/xlsx.
        index_col : None or int
            Index column of the :class:`pandas.DataFrame`.

        Tip
        ---
        Refer to the `Bwaise system <https://github.com/QSD-Group/EXPOsan/tree/main/exposan/bwaise/data>`_
        in the `Exposan` repository for a sample file.
        '''
        if isinstance(path_or_dict, str):
            if not (path_or_dict.endswith('.xls') or path_or_dict.endswith('.xlsx')):
                raise ValueError('Only Excel files ends with ".xlsx" or ".xls" can be interpreted.')

            data_file = pd.ExcelFile(path_or_dict, engine='openpyxl')

            for sheet_name in data_file.sheet_names:
                data = data_file.parse(sheet_name, index_col=index_col)
                cls._load_from_df(sheet_name, data)
        else:
            for k, v in path_or_dict.items():
                cls._load_from_df(k, v)

    @property
    def source(self):
        '''
        [:class:`ImpactItem`] If provided, all attributions and properties of this
        impact item will be copied from the provided source.
        ID of the impact item can be provided instead of the object.
        '''
        return self._source
    @source.setter
    def source(self, i):
        if not isinstance(i, ImpactItem):
            if isinstance(i, str):
                i = self.get_item(i)
            elif not i:
                i = None
            else:
                raise ValueError('`source` can only be an `ImpactItem` or its ID, '
                                 f'not {type(i).__name__}.')
        self._source = i

    @property
    def ID(self):
        '''[str] ID of this item.'''
        return self._ID
    @ID.setter
    def ID(self, ID):
        self._ID = ID

    @property
    def functional_unit(self):
        '''[str] Functional unit of this item.'''
        return check_source(self, True, False)._functional_unit.units
    @functional_unit.setter
    def functional_unit(self, i):
        check_source(self, True)._functional_unit = auom(i)

    @property
    def indicators(self):
        ''' [tuple] Impact indicators associated with this item.'''
        return tuple(ImpactIndicator.get_all_indicators(True)[i]
                     for i in self.CFs.keys())

    @property
    def price(self):
        '''Price of this item per functional unit.'''
        return check_source(self, True, False)._price
    @price.setter
    def price(self, price, unit=''):
        check_source(self, True)._update_price(price, unit)

    @property
    def CFs(self):
        '''[dict] Characterization factors of this item for different impact indicators.'''
        if self.source:
            return self.source._CFs.copy()
        else:
            return self._CFs
    @CFs.setter
    def CFs(self, indicator, CF_value, CF_unit=''):
        check_source(self, True).add_indicator(indicator, CF_value, CF_unit)

    @property
    def registered(self):
        '''[bool] If this impact item is registered in the record.'''
        data = self.registry.data.get(self.ID)
        return True if data is self else False


# %%

class StreamImpactItem(ImpactItem):
    '''
    A class for calculation of environmental impacts associated with streams
    (e.g., chemical inputs, emissions).

    Parameters
    ----------
    ID : str
        ID of the item. If no ID is provided but its `linked_stream` is provided,
        the ID will be set as ID of the `linked_stream` with a suffix '_item'
    linked_stream : :class:`SanStream`
        The associated :class:`SanStream` for environmental impact calculation.
    source : :class:`StreamImpactItem`
        If provided, all attributions and properties of this
        :class:`StreamImpactItem` will be copied from the provided source.
    indicator_CFs : kwargs
        ImpactIndicators and their characteriziation factors (CFs).

    Tip
    ---
    For environmental impacts associated with construction and transportation,
    use :class:`ImpactItem` instead.

    Examples
    --------
    Refer to :class:`ImpactItem` for general features.
    Below is about the additional features for :class:`StreamImpactItem`.

    Assume we want to account for the globalwarming potential for methane:

    >>> # Make impact indicators
    >>> import qsdsan as qs
    >>> GWP = qs.ImpactIndicator('GlobalWarming', alias='GWP', unit='kg CO2-eq')
    >>> FEC = qs.ImpactIndicator('FossilEnergyConsumption', alias='FEC', unit='MJ')
    >>> # Make an stream impact item
    >>> methane_item = qs.StreamImpactItem('methane_item', GWP=28)
    >>> methane_item.show()
    StreamImpactItem: methane_item [per kg]
    Linked to       : None
    Price           : None USD
    ImpactIndicators:
                               Characterization factors
    GlobalWarming (kg CO2-eq)                        28
    >>> # Make a stream and link the stream to the impact item
    >>> cmps = qs.utils.load_example_cmps()
    >>> qs.set_thermo(cmps)
    >>> methane = qs.SanStream('methane', Methane=1, units='kg/hr',
    ...                        stream_impact_item=methane_item)
    >>> methane_item.show()
    StreamImpactItem: methane_item [per kg]
    Linked to       : methane
    Price           : None USD
    ImpactIndicators:
                               Characterization factors
    GlobalWarming (kg CO2-eq)                        28

    We can make copies of the impact item, and link it to the original one.

    >>> methane2 = methane.copy('methane2')
    >>> methane_item2 = methane_item.copy('methane_item2', stream=methane2,
    ...                                   set_as_source=True)
    >>> methane_item2.CFs['GlobalWarming']
    28
    >>> methane_item.CFs['GlobalWarming'] = 1
    >>> methane_item2.CFs['GlobalWarming']
    1

    We can also add or remove impact indicators.

    >>> methane_item.remove_indicator('GlobalWarming')
    The impact indicator "GlobalWarming" has been removed.
    >>> methane_item2.show()
    StreamImpactItem: methane_item2 [per kg]
    Linked to       : methane2
    Source          : methane_item
    Price           : None USD
    ImpactIndicators:
     None
    >>> methane_item.add_indicator('GlobalWarming', 28)
    >>> methane_item2.show()
    StreamImpactItem: methane_item2 [per kg]
    Linked to       : methane2
    Source          : methane_item
    Price           : None USD
    ImpactIndicators:
                               Characterization factors
    GlobalWarming (kg CO2-eq)                        28
    '''

    __slots__ = ('_ID', '_linked_stream', '_functional_unit', '_CFs', '_source')

    def __init__(self, ID='', linked_stream=None, source=None, **indicator_CFs):

        self._linked_stream = None
        self.linked_stream = linked_stream
        self._register(ID)

        if not ID and linked_stream:
            ID = self.linked_stream.ID + '_item'
        self._ID = ID

        if source:
            self.source = source
        else:
            self._source = None
            self._functional_unit = auom('kg')
            self._CFs = {}
            for CF, value in indicator_CFs.items():
                try:
                    CF_value, CF_unit = value # unit provided for CF
                    self.add_indicator(CF, CF_value, CF_unit)
                except Exception as e:
                    if 'unpack' in str(sys.exc_info()[1]):
                        self.add_indicator(CF, value)
                    else:
                        raise e


    def __repr__(self):
        return f'<StreamImpactItem: {self.ID}>'


    def show(self):
        '''Show basic information about this :class:`StreamImpactItem` object.'''
        info = f'StreamImpactItem: {self.ID} [per {self.functional_unit}]'
        if self.linked_stream:
            info += f'\nLinked to       : {self.linked_stream}'
        else:
            info += '\nLinked to       : None'
        if self.source:
            info += f'\nSource          : {self.source.ID}'
        info += f'\nPrice           : {f_num(self.price)} {currency}'
        info += '\nImpactIndicators:'
        print(info)
        if len(self.CFs) == 0:
            print(' None')
        else:
            index = pd.Index((i.ID+' ('+i.unit+')' for i in self.indicators))
            df = pd.DataFrame({
                'Characterization factors': tuple(self.CFs.values())
                },
                index=index)
            # print(' '*18+df.to_string().replace('\n', '\n'+' '*18))
            print(df.to_string())


    _ipython_display_ = show


    def copy(self, new_ID='', stream=None, set_as_source=False):
        '''
        Return a new :class:`StreamImpactItem` object with the same settings.

        Parameters
        ----------
        new_ID : str
            ID of the new stream impact item.
        stream : :class:`~.SanStream`
            Linked stream to the copy.
        set_as_source : bool
            Whether to set the original impact item as the source.
        '''

        new = StreamImpactItem.__new__(StreamImpactItem)
        new.ID = new_ID
        new._linked_stream = None # initiate this attribute

        if set_as_source:
            if self.source:
                new.source = self.source
            else:
                new.source = self
        else:
            new = copy_attr(new, self, skip=('_ID', '_linked_stream', '_source'))

        if stream:
            stream.stream_impact_item = new
            if not new_ID:
                new.ID = f'{stream.ID}_item'

        return new

    __copy__ = copy

    @property
    def source(self):
        '''
        [:class:`StreamImpactItem`] If provided, all attributions and properties of this
        :class:`StreamImpactItem` will be copied from the provided source.
        ID of the impact item can be provided instead of the object.

        .. note::

            Since the price is copied from the price of the `linked_stream`, it
            can be different from the source.
        '''
        return self._source
    @source.setter
    def source(self, i):
        if isinstance(i, str):
            i = self.get_item(i)

        if not isinstance(i, StreamImpactItem):
            if not i:
                i = None
            else:
                raise ValueError('`source` can only be a StreamImpactItem, ' \
                                 f'not a {type(i).__name__}.')
        self._source = i


    @property
    def linked_stream(self):
        '''
        [:class:`SanStream`] The associated :class:`SanStream` for environmental impact calculation,
        can be set by either the :class:`SanStream` object or its ID.
        '''
        return self._linked_stream

    @linked_stream.setter
    def linked_stream(self, new_s):
        if new_s and not isinstance(new_s, SanStream):
            if isinstance(new_s, str):
                try:
                    new_s = getattr(SanStream.registry, new_s)
                except:
                    try:
                        new_s = getattr(WasteStream.registry, new_s)
                    except:
                        raise ValueError(f'The ID "{new_s}" not found in registry.')
            else:
                raise TypeError('`linked_stream` must be a `SanStream` or '
                                f'the ID of a `SanStream`, not {type(new_s).__name__}.')

        if self._linked_stream:
            old_s = self._linked_stream
            self._linked_stream.stream_impact_item = None
            warn(f'`StreamImpactItem` {self.ID} is unlinked from {old_s.ID} and ' \
                 f'linked to {new_s.ID}.', stacklevel=2)
        if new_s:
            if hasattr(self, '_ID'):
                if new_s.stream_impact_item and new_s.stream_impact_item.ID != self.ID:
                    msg = f'The original `StreamImpactItem` linked to stream {new_s} ' \
                        f'is replaced with {self}.'
                    warn(message=msg, stacklevel=2)
            new_s._stream_impact_item = self
        self._linked_stream = new_s

    @property
    def ID(self):
        '''
        [str] ID of the item. If no ID is provided but its `linked_stream` is provided,
        the ID will be set as ID of the `linked_stream` with a suffix '_item'.
        '''
        return self._ID
    @ID.setter
    def ID(self, i):
        self._ID = i

    @property
    def functional_unit(self):
        '''[str] Functional unit of the item, set to 'kg'.'''
        return auom('kg')

    @property
    def price(self):
        '''[float] Price of the linked stream.'''
        if self.linked_stream:
            return self.linked_stream.price
        else:
            return 0.


# -*- coding: utf-8 -*-
'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Joy Cheung <joycheung1994@gmail.com>

Part of this module is based on the Thermosteam package:
https://github.com/BioSTEAMDevelopmentGroup/thermosteam

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''

# import thermosteam as tmo
from . import Components
from .utils.loading import load_data
from .utils.parse import get_stoichiometric_coeff
from thermosteam.utils import chemicals_user, read_only
from sympy import symbols, Matrix, lambdify
from sympy.parsing.sympy_parser import parse_expr
import numpy as np
import pandas as pd
    
__all__ = ('Process', 'Processes', 'CompiledProcesses', )

class UndefinedProcess(AttributeError):
    '''AttributeError regarding undefined Component objects.'''
    def __init__(self, ID):
        super().__init__(repr(ID))
        
#%%
@chemicals_user        
class Process():
    """
    Create a :class:`Process` object which defines a stoichiometric process and its kinetics.
    A :class:`Process` object is capable of reacting the component flow rates of a :class:`WasteStream`
    object.

    Parameters
    ----------
    ID : str
        A unique identification.
    reaction : dict, str, or numpy.ndarray
        A dictionary of stoichiometric coefficients with component IDs as 
        keys, or a numeric array of stoichiometric coefficients, or a string 
        of a stoichiometric equation written as: 
        i1 R1 + ... + in Rn -> j1 P1 + ... + jm Pm.
        Stoichiometric coefficients can be symbolic or numerical. 
        Unknown stoichiometric coefficients to solve for should be expressed as "?".
    ref_component : str
        ID of the reference :class:`Component` object of the process rate.
    rate_equation : str, optional
        The kinetic rate equation of the process. The default is None.
    components=None : class:`CompiledComponents`, optional
        Components corresponding to each entry in the stoichiometry array, 
        defaults to thermosteam.settings.chemicals. 
    conserved_for : tuple[str], optional
        Materials subject to conservation rules, must be an 'i\_' attribute of
        the components. The default is ("COD", "N", "P", "charge").
    parameters : Iterable[str], optional
        Symbolic parameters in stoichiometry coefficients and/or rate equation. 
        The default is None.

    Examples
    --------
    None.

    """

    def __init__(self, ID, reaction, ref_component, rate_equation=None, components=None, 
                 conserved_for=('COD', 'N', 'P', 'charge'), parameters=None):

        self._ID = ID
        self._stoichiometry = []
        self._components = self._load_chemicals(components)
        self._ref_component = ref_component
        self._conserved_for = conserved_for
        self._parameters = {p: symbols(p) for p in parameters}
        
        self._stoichiometry = get_stoichiometric_coeff(reaction, self._ref_component, self._components, self._conserved_for, self._parameters)
        self._parse_rate_eq(rate_equation)
                
    def get_conversion_factors(self, as_matrix=False):        
        '''
        return conversion factors (i.e., the 'i\_' attributes of the components) 
        as a numpy.ndarray or a SymPy Matrix.
        '''
        if self._conservation_for:
            cmps = self._components
            arr = getattr(cmps, 'i_'+self._conservation_for[0])
            for c in self._conservation_for[1:]:
                arr = np.vstack((arr, getattr(cmps, 'i_'+c)))
            if as_matrix: return Matrix(arr.tolist())
            return arr
        else: return None
                    
    def check_conservation(self, tol=1e-8):
        '''check conservation of materials given numerical stoichiometric coefficients. '''
        isa = isinstance
        if isa(self._stoichiometry, np.ndarray):
            ic = self.get_conversion_factors()
            v = self._stoichiometry
            ic_dot_v = ic @ v
            conserved_arr = np.isclose(ic_dot_v, np.zeros(ic_dot_v.shape), atol=tol)
            if not conserved_arr.all(): 
                materials = self._conserved_for
                unconserved = [(materials[i], ic_dot_v[i]) for i, conserved in enumerate(conserved_arr) if not conserved]
                raise RuntimeError("The following materials are unconserved by the "
                                   "stoichiometric coefficients. A positive value "
                                   "means the material is created, a negative value "
                                   "means the material is destroyed:\n "
                                   + "\n ".join([f"{material}: {value:.2f}" for material, value in unconserved]))
        else: 
            raise RuntimeError("Can only check conservations with numerical "
                               "stoichiometric coefficients.")
    
    def reverse(self):
        '''reverse the process as to flip the signs of stoichiometric coefficients of all components.'''
        if isinstance(self._stoichiometry, np.ndarray):
            self._stoichiometry = -self._stoichiometry
        else:
            self._stoichiometry = [-v for v in self._stoichiometry]
        self._rate_equation = -self._rate_equation
        
    @property
    def ID(self):
        '''[str] A unique identification'''
        return self._ID
    
    @property
    def ref_component(self):
        '''
        [str] ID of the reference component
        
        .. note::
    
            When a new value is assigned, all stoichiometric coefficient will be 
            normalized so that the new stoichiometric coefficient of the new reference
            component is 1 or -1. The rate equation will also be updated automatically.
        '''
        return getattr(self._components, self._ref_component)    
    @ref_component.setter
    def ref_component(self, ref_cmp):
        if ref_cmp: 
            self._ref_component = ref_cmp
            self._normalize_stoichiometry(ref_cmp)
            self._normalize_rate_eq(ref_cmp)

    @property
    def conserved_for(self):
        '''
        [tuple] Materials subject to conservation rules, must have corresponding 
        'i\_' attributes for the components
        '''
        return self._conserved_for
    @conserved_for.setter
    def conserved_for(self, materials):
        self._conserved_for = materials
    
    @property
    def parameters(self):
        '''[dict] Symbolic parameters in stoichiometric coefficients and rate equation.'''
        return self._parameters
    
    def append_parameters(self, *new_pars):
        '''append new symbolic parameters'''
        for p in new_pars:
            self._parameters[p] = symbols(p)
    
    def set_parameters(self, **parameters):
        '''
        Set values for symbolic stoichiometric and/or kinetic parameters.

        Parameters
        ----------
        **parameters : float, int, or str
            Parameters and values.
        '''
        self._parameters.update(parameters)
    
    @property
    def stoichiometry(self):
        '''[dict] Non-zero stoichiometric coefficients.'''
        allcmps = dict(zip(self._components.IDs, self._stoichiometry))
        active_cmps = {k:v for k,v in allcmps.items() if v != 0}        
        if isinstance(self._stoichiometry, list):
            active_cmps = {k:v.subs(self._parameters) for k,v in active_cmps.items()}
        return active_cmps
        
    @property
    def rate_equation(self):
        '''
        [SymPy expression] Kinetic rate equation of the process. Also the rate in
        which the reference component is reacted or produced in the process.
        '''
        return self._rate_equation.subs(self._parameters)
    
    def _parse_rate_eq(self, eq):
        cmpconc_symbols = {c: symbols(c) for c in self._components.IDs}
        self._rate_equation = parse_expr(eq, {**cmpconc_symbols, **self._parameters})
    
    def _normalize_stoichiometry(self, new_ref):
        isa = isinstance
        factor = abs(self._stoichiometry[self._components._index[new_ref]])
        if isa(self._stoichiometry, np.ndarray):
            self._stoichiometry /= factor
        elif isa(self._stoichiometry, list):
            self._stoichiometry = [v/factor for v in self._stoichiometry]
    
    def _normalize_rate_eq(self, new_ref):
        factor = self._stoichiometry[self._components._index[new_ref]]
        self._rate_equation *= factor
    
    def show(self):
        info = f"Process: {self.ID}"
        header = '\n[stoichiometry] '
        section = []
        for cmp, stoichio in tuple(zip(self._components.IDs, self._stoichiometry)):
            if stoichio != 0:
                if isinstance(stoichio, (int, float)): line = f"{cmp}: {stoichio:.3g}"
                else: line = f"{cmp}: {stoichio.evalf(n=3)}"
                section.append(line)
        info += header + ("\n" + 16*" ").join(section)
        info += '\n[reference]     ' + f"{self.ref_component.ID}"
        line = '\n[rate equation] ' + f"{self._rate_equation}"
        if len(line) > 47: line = line[:47] + '...'
        info += line
        header = '\n[parameters]    '
        section = []
        for k,v in self.parameters.items():
            if isinstance(v, (int, float)): line = f"{k}: {v:.3g}"
            else: line = f"{k}: {v}"
            section.append(line)
        info += header + ("\n" + 16*" ").join(section)
        print(info)

    _ipython_display_ = show
        
    # TODO: return a copy
    def copy():
        pass
    
#%%
setattr = object.__setattr__
@chemicals_user
class Processes():
    """
    Create a :class:`Processes` object that contains :class:`Process` objects as attributes.

    Parameters
    ----------
    processes : Iterable[:class:`Process`]

    Examples
    --------
    None.

    """
    
    def __new__(cls, processes):

        self = super().__new__(cls)
        #!!! add function to detect duplicated processes
        setfield = setattr
        for i in processes:
            setfield(self, i.ID, i)
        return self
    
    # def __getnewargs__(self):
    #     return(tuple(self),)
    
    def __setattr__(self, ID, process):
        raise TypeError("can't set attribute; use <Processes>.append or <Processes>.extend instead")
    
    def __setitem__(self, ID, process):
        raise TypeError("can't set attribute; use <Processes>.append or <Processes>.extend instead")
    
    def __getitem__(self, key):
        """
        Return a :class:`Process` object or a list of :class:`Process` objects.
        
        Parameters
        ----------
        key : Iterable[str] or str
              Process IDs.
        
        """
        dct = self.__dict__
        try:
            if isinstance(key, str):
                return dct[key]
            else:
                return [dct[i] for i in key]
        except KeyError:
            raise KeyError(f"undefined process {key}")
    
    def copy(self):
        """Return a copy."""
        copy = object.__new__(Processes)
        for proc in self: setattr(copy, proc.ID, proc)
        return copy
    
    def append(self, process):
        """Append a :class:`Process`."""
        if not isinstance(process, Process):
            raise TypeError("only 'Process' objects can be appended, "
                           f"not '{type(process).__name__}'")
        ID = process.ID
        if ID in self.__dict__:
            raise ValueError(f"{ID} already defined in processes")
        setattr(self, ID, process)
    
    def extend(self, processes):
        """Extend with more :class:`Process` objects."""
        if isinstance(processes, Processes):
            self.__dict__.update(processes.__dict__)
        else:
            for process in processes: self.append(process)
    
    def subgroup(self, IDs):
        """
        Create a new subgroup of processes.
        
        Parameters
        ----------
        IDs : Iterable[str]
              Process IDs.
              
        """
        return Process([getattr(self, i) for i in IDs])
    
    def compile(self, skip_checks=False):
        '''Cast as a :class:`CompiledProcesses` object.'''
        processes = tuple(self)
        setattr(self, '__class__', CompiledProcesses)
        try: self._compile(processes, skip_checks)
        except Exception as error:
            setattr(self, '__class__', Processes)
            setattr(self, '__dict__', {i.ID: i for i in processes})
            raise error
            
    # kwarray = array = index = indices = must_compile
        
    def show(self):
        print(self)
    
    _ipython_display_ = show
    
    def __len__(self):
        return len(self.__dict__)
    
    def __contains__(self, process):
        if isinstance(process, str):
            return process in self.__dict__
        elif isinstance(process, Process):
            return process in self.__dict__.values()
        else: # pragma: no cover
            return False
    
    def __iter__(self):
        yield from self.__dict__.values()
    
    def __repr__(self):
        return f"{type(self).__name__}([{', '.join(self.__dict__)}])"
    
    _default_data = None
    
    @classmethod
    def load_from_file(cls, path='', components=None, 
                       conserved_for=('COD', 'N', 'P', 'charge'), parameters=None,
                       use_default_data=False, store_data=False, compile=True):
        """
        Create :class:`CompiledProcesses` object from a table of process IDs, stoichiometric 
        coefficients, and rate equations stored in a .tsv, .csv, or Excel file. 

        Parameters
        ----------
        path : str, optional
            File path.
        components : :class:`CompiledComponents`, optional
            Components corresponding to the columns in the stoichiometry matrix, 
            defaults to thermosteam.settings.chemicals. The default is None.
        conserved_for : tuple[str], optional
            Materials subject to conservation rules, must have corresponding 'i\_' 
            attributes for the components. The default is ('COD', 'N', 'P', 'charge').
        parameters : Iterable[str], optional
            Symbolic parameters in waste. The default is None.
        use_default_data : bool, optional
            Whether to use default data. The default is False.
        store_data : bool, optional
            Whether to store the file as default data. The default is False.
        compile : bool, optional
            Whether to compile processes. The default is True.

        .. note::
    
            [1] First column of the table should be process IDs, followed by stoichiometric 
            coefficient matrix with corresponding component IDs as column names, and rate 
            equations as the last column. 
            
            [2] Entries of stoichiometric coefficients can be symbolic or numerical. 
            Blank cells are considered zero.
            
            [3] Unknown stoichiometric coefficients to solve for using conservation 
            rules should be uniformly written as '?'. 
            
            [4] For each process, the first component with stoichiometric coefficient
            of -1 or 1 is considered the reference component. If none of the components
            has -1 or 1 stoichiometric coefficient, the first component with non-zero
            coefficient is considered the reference.
        """
        if use_default_data and cls._default_data is not None:
            data = cls._default_data
        else:
            data = load_data(path=path, index_col=None, na_values=0)
        
        cmp_IDs = data.columns[1:-1]
        data.dropna(how='all', subset=cmp_IDs, inplace=True)
        new = cls(())
        for i, proc in data.iterrows():
            ID = proc[0]
            stoichio = proc[1:-1]
            if pd.isna(proc[-1]): rate_eq = None
            else: rate_eq = proc[-1]
            stoichio = stoichio[-pd.isna(stoichio)].to_dict()
            for k,v in stoichio.items():
                try: stoichio[k] = float(v)
                except: continue
            ref = [k for k,v in stoichio.items() if v in (-1, 1)]
            if len(ref) == 0: ref = list(stoichio.keys())[0]                
            else: ref = ref[0]            
            process = Process(ID, stoichio, 
                              ref_component=ref, 
                              rate_equation=rate_eq,
                              components=components,
                              conserved_for=conserved_for,
                              parameters=parameters)
            new.append(process)

        if store_data:
            cls._default_data = data
        
        if compile: new.compile()
        return new
        
            
#%%
@read_only(methods=('append', 'extend', '__setitem__'))
class CompiledProcesses(Processes):
    """
    Create a :class:`CompiledProcesses` object that contains :class:`Process` objects as attributes.

    Parameters
    ----------
    processes : Iterable[:class:`Process`]

    Examples
    --------
    None.

    """
    
    _cache = {}
    
    def __new__(cls, processes):
        cache = cls._cache
        processes = tuple(processes)
        if processes in cache:
            self = cache[processes]
        else:
            self = object.__new__(cls)
            setfield = setattr
            for i in processes:
                setfield(self, i.ID, i)
            self._compile()
            cache[processes] = self
        return self

    # def __dir__(self):
    #     pass
    
    def compile(self, skip_checks=False):
        """Do nothing, :class:`CompiledProcesses` objects are already compiled."""
        pass
    
    def _compile(self, processes, skip_checks=False):
        dct = self.__dict__
        isa = isinstance
        tuple_ = tuple
        # processes = tuple_(dct.values())
        missing_rate_proc = []
        for process in processes:
            if process.rate_equation == None: 
                missing_rate_proc.append(process.ID)
        if not skip_checks and len(missing_rate_proc) > 0: 
            raise RuntimeError(f"The following processes are missing rate equations:"
                               f"{missing_rate_proc}")        
        IDs = tuple_([i.ID for i in processes])
        size = len(IDs)
        index = tuple_(range(size))
        dct['tuple'] = processes
        dct['size'] = size
        dct['IDs'] = IDs
        dct['_index'] = index = dict(zip(IDs, index))
        cmps = Components([cmp for i in processes for cmp in i._components])
        cmps.compile()
        dct['_components'] = cmps
        M_stch = []
        params = {}
        rate_eqs = tuple_([i._rate_equation for i in processes])
        all_numeric = True
        for i in processes:
            stch = [0]*cmps.size
            params.update(i._parameters)
            if all_numeric and isa(i._stoichiometry, (list, tuple)): all_numeric = False
            for cmp, coeff in i.stoichiometry.items():
                stch[cmps._index[cmp]] = coeff
            M_stch.append(stch)
        dct['_parameters'] = params
        if all_numeric: M_stch = np.asarray(M_stch)
        dct['_stoichiometry'] = M_stch
        dct['_rate_equations'] = rate_eqs
        dct['_production_rates'] = list(Matrix(M_stch).T * Matrix(rate_eqs))
        
    @property
    def parameters(self):
        '''[dict] All symbolic stoichiometric and kinetic parameters.'''
        return self._parameters
    
    @property
    def stoichiometry(self):
        '''[pandas.DataFrame] Stoichiometric coefficients.'''
        stoichio = self._stoichiometry
        isa = isinstance
        v_params = {k:v for k,v in self._parameters.items() if isa(v, (float, int))}
        if isa(stoichio, list) and len(v_params) > 0:
            stoichio_vals = []
            for row in stoichio:
                stoichio_vals.append([v.subs(v_params) if not isa(v, (float, int)) else v for v in row])
            return pd.DataFrame(stoichio_vals, index=self.IDs, columns=self._components.IDs)
        else: return pd.DataFrame(stoichio, index=self.IDs, columns=self._components.IDs)
    
    @property
    def rate_equations(self):
        '''[pandas.DataFrame] Rate equations.'''
        rate_eqs = [eq.subs(self._parameters) for eq in self._rate_equations]
        return pd.DataFrame(rate_eqs, index=self.IDs, columns=('rate_equation',))
    
    # @property
    # def rate_equations_function(self):
    #     '''
    #     [function] A function object of the rate equations. When evaluated, 
    #     returns a list of process rates.
    #     '''
    #     return self._rate_equations_function
    
    @property
    def production_rates(self):
        '''[pandas.DataFrame] The rates of production of the components.'''
        rates = [r.subs(self._parameters) for r in self._production_rates]
        return pd.DataFrame(rates, index=self._components.IDs, columns=('rate_of_production',))
    
    # @property
    # def production_rates_function(self):
    #     '''
    #     [function] A function object of the components' rates of production. 
    #     When evaluated, returns a list of production rates.
    #     '''
    #     args = list(symbols(self._components.IDs)) + [v for k,v in self._parameters.items() if k == str(v)]
    #     return [lambdify(args, r) for r in self._production_rates]
    
    def subgroup(self, IDs):
        '''Create a new subgroup of :class:`CompiledProcesses` objects.'''
        processes = self[IDs]
        new = Processes(processes)
        new.compile()
        return new
    
    def index(self, ID):
        '''Return index of specified process.'''
        try: return self._index[ID]
        except KeyError:
            raise UndefinedProcess(ID)

    def indices(self, IDs):
        '''Return indices of multiple processes.'''
        try:
            dct = self._index
            return [dct[i] for i in IDs]
        except KeyError as key_error:
            raise UndefinedProcess(key_error.args[0])
    
    def __contains__(self, process):
        if isinstance(process, str):
            return process in self.__dict__
        elif isinstance(process, Process):
            return process in self.tuple
        else: # pragma: no cover
            return False
    
    def __len__(self):
        return self.size
    
    def __iter__(self):
        return iter(self.tuple)
        
    def copy(self):
        '''Return a copy.'''
        copy = Processes(self.tuple)
        copy.compile()
        return copy    
    
    def set_parameters(self, **parameters):
        '''Set values to stoichiometric and/or kinetic parameters.'''
        self._parameters.update(parameters)
    
    def __repr__(self):
        return f"{type(self).__name__}([{', '.join(self.IDs)}])"


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''


# %%

import pandas as pd
from thermosteam.utils import registered
from . import currency, ImpactItem
from .utils import (
    auom, copy_attr,
    format_number as f_num
    )

__all__ = ('Construction',)


@registered(ticket_name='Constr')
class Construction:
    '''
    Construction activity for cost and environmental impact calculations.

    Parameters
    ----------
    ID : str
        ID of this construction activity.
    item : :class:`ImpactItem`
        Impact item associated with this consturction activity.
    quantity : float
        Quantity of the impact item involved in this construction activity.
    lifetime : float
        Lifetime of the constructed item.
    lifetime_unit : str
        Unit of the lifetime.

    Examples
    --------
    >>> import qsdsan as qs
    >>> # Make impact indicators
    >>> GWP = qs.ImpactIndicator('GlobalWarming', alias='GWP', unit='kg CO2-eq')
    >>> FEC = qs.ImpactIndicator('FossilEnergyConsumption', alias='FEC', unit='MJ')
    >>> # Make impact item
    >>> Steel = qs.ImpactItem('Steel', 'kg', GWP=2.55, FEC=0.5)
    >>> # Make a construction activity that uses 100 g of steel
    >>> steel_100_g = qs.Construction('steel_100_g', item=Steel, quantity=100,
    ...                               quantity_unit='g')
    >>> steel_100_g.show()
    Construction : steel_100_g
    Impact item  : Steel
    Lifetime     : None yr
    Quantity     : 0.1 kg
    Total cost   : None USD
    Total impacts:
                                  Impacts
    GlobalWarming (kg CO2-eq)       0.255
    FossilEnergyConsumption (MJ)     0.05
    '''

    __slots__ = ('_ID', '_item', '_quantity', '_lifetime')

    def __init__(self, ID='', item=None, quantity=0., quantity_unit='',
                 lifetime=None, lifetime_unit='yr'):
        self._register(ID)
        self.item = item
        self._update_quantity(quantity, quantity_unit)
        self._lifetime = None
        if lifetime:
            self._lifetime = auom(lifetime_unit).convert(lifetime, 'yr')

    def _update_quantity(self, quantity=0., quantity_unit=''):
        if not quantity_unit or quantity_unit == self.item.functional_unit:
            self._quantity = float(quantity)
        else:
            converted = auom(quantity_unit).convert(float(quantity), self.item.functional_unit)
            self._quantity = converted

    def __repr__(self):
        return f'<Construction: {self.ID}>'

    def show(self):
        '''Show basic information about this :class:`Construction` object.'''
        item = self.item
        impacts = self.impacts
        info = f'Construction : {self.ID}'
        info += f'\nImpact item  : {item.ID}'
        info += f'\nLifetime     : {f_num(self.lifetime)} yr'
        info += f'\nQuantity     : {f_num(self.quantity)} {item.functional_unit}'
        info += f'\nTotal cost   : {f_num(self.cost)} {currency}'
        info += '\nTotal impacts:'
        print(info)
        if len(impacts) == 0:
            print(' None')
        else:
            index = pd.Index((i.ID+' ('+i.unit+')' for i in self.indicators))
            df = pd.DataFrame({
                'Impacts': tuple(self.impacts.values())
                },
                index=index)
            # print(' '*15+df.to_string().replace('\n', '\n'+' '*15))
            print(df.to_string())

    _ipython_display_ = show

    def copy(self, new_ID=''):
        new = Construction.__new__(Construction)
        new.__init__(new_ID)
        new = copy_attr(new, self, skip=('_ID',))
        return new

    __copy__ = copy

    @property
    def lifetime(self):
        '''[float] Lifetime of this construction activity.'''
        return self._lifetime
    @lifetime.setter
    def lifetime(self, lifetime, unit='yr'):
        if lifetime is None:
            self.lifetime = lifetime
        else:
            self._lifetime = auom(unit).convert(lifetime, 'yr')

    @property
    def item(self):
        '''[:class:`ImpactItem`] The impact item associated with this construction activity.'''
        return self._item
    @item.setter
    def item(self, i):
        if not i:
            i = None
        elif isinstance(i, str):
            i = ImpactItem.get_item(i)
        elif not isinstance(i, ImpactItem):
            raise TypeError('Only `ImpactItem` or the ID of `ImpactItem` can be set, '
                            f'not {type(i).__name__}.')
        self._item = i

    @property
    def indicators(self):
        ''' [tuple] Impact indicators associated with the construction item.'''
        return self.item.indicators

    @property
    def quantity(self):
        '''[float] Quantity of this construction item.'''
        return self._quantity
    @quantity.setter
    def quantity(self, quantity, unit=''):
        self._update_quantity(quantity, unit)

    @property
    def price(self):
        '''[float] Unit price of the item.'''
        return self.item.price

    @property
    def cost(self):
        '''[float] Total cost of this construction item.'''
        return self.quantity*self.price

    @property
    def impacts(self):
        '''[dict] Total impacts of this construction activity over its lifetime.'''
        impacts = {}
        for indicator, CF in self.item.CFs.items():
            impacts[indicator] = self.quantity*CF
        return impacts


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''


# %%

from warnings import warn
from thermosteam.utils import registered
from .utils import parse_unit, load_data

__all__ = ('ImpactIndicator', )


@registered(ticket_name='ind')
class ImpactIndicator:
    '''
    To handle different impact indicators in life cycle assessment.

    Parameters
    ----------
    ID : str
        ID of this impact indicator.
    alias : str
        Alternative ID of this impact indicator.

        .. note::

            "synonym" was used bfore v0.2.2 it is still supported, but may be
            removed in the future.

    method : str
        Impact assessment method, e.g., 'TRACI'.
    category : str
        Category of this impact indicator, e.g., 'human health'.
    unit : str
        Unit of this impact indicator, e.g., 'kg CO2-eq'.
    description : str
        Supplementary explanation.

    Examples
    --------
    Make an impact indicator for global warming potential.

    >>> import qsdsan as qs
    >>> GWP = qs.ImpactIndicator('GlobalWarming', method='TRACI',
    ...                          category='environmental impact',
    ...                          unit='kg CO2-eq',
    ...                          description='Effect of climate change measured as \
    ...                                      global warming potential.')

    See relevant information.

    >>> GWP.show()
    ImpactIndicator: GlobalWarming as kg CO2-eq
     Alias      : None
     Method     : TRACI
     Category   : environmental impact
     Description: Effect of climate change ...
    >>> # Add an alias
    >>> GWP.alias = 'GWP'
    >>> GWP.show()
    ImpactIndicator: GlobalWarming as kg CO2-eq
     Alias      : GWP
     Method     : TRACI
     Category   : environmental impact
     Description: Effect of climate change ...
    >>> # Add another impact indicator
    >>> FEC = qs.ImpactIndicator('FossilEnergyConsumption', alias='FEC', unit='MJ')
    >>> # Get all impact indicators
    >>> qs.ImpactIndicator.get_all_indicators()
    {'GlobalWarming': <ImpactIndicator: GlobalWarming>,
     'FossilEnergyConsumption': <ImpactIndicator: FossilEnergyConsumption>}

    Manage the registry.

    >>> GWP.deregister()
    The impact indicator "GlobalWarming" has been removed from the registry.
    >>> qs.ImpactIndicator.get_all_indicators()
    {'FossilEnergyConsumption': <ImpactIndicator: FossilEnergyConsumption>}
    >>> GWP.register()
    The impact indicator "GlobalWarming" has been added to the registry.
    >>> qs.ImpactIndicator.get_all_indicators()
    {'FossilEnergyConsumption': <ImpactIndicator: FossilEnergyConsumption>,
     'GlobalWarming': <ImpactIndicator: GlobalWarming>}
    >>> qs.ImpactIndicator.clear_registry()
    All impact indicators have been removed from registry.
    >>> qs.ImpactIndicator.get_all_indicators()
    {}
    '''

    __slots__ = ('_ID', '_alias', '_method', '_category', '_unit', '_ureg_unit',
                 '_unit_remaining', '_description')

    def __init__(self, ID='', alias='', method='', category='', unit='', description='',
                 **kwargs):

        self._register(ID)
        self.alias = alias

        self._unit = str(unit)
        self._ureg_unit, self._unit_remaining = parse_unit(unit)
        self._method = method
        self._category = category
        self._description = description

        if 'synonym' in kwargs.keys():
            synonym = kwargs['synonym']
            if (not alias or str(alias)=='nan'):
                raise DeprecationWarning('`synonym` has been changed to `alias` for qsdsan v0.2.2 and above.')
                alias = synonym
            else:
                raise DeprecationWarning('`synonym` has been changed to `alias` for qsdsan v0.2.2 and above, ' \
                                         f'the given `synonym` "{synonym}" is ignored as `alias` "{alias}" is provided.')


    def __repr__(self):
        return f'<ImpactIndicator: {self.ID}>'

    def show(self):
        '''Show basic information about this impact indicator.'''
        if self.unit:
            info = f'ImpactIndicator: {self.ID} as {self.unit}'
        else:
            info = f'ImpactIndicator: {self.ID}'

        alias = self.alias if self.alias else 'None'
        line =   f'\n Alias      : {alias}'
        if len(line) > 40: line = line[:40] + '...'
        info += line

        info += f'\n Method     : {self.method or None}'
        info += f'\n Category   : {self.category or None}'
        line =  f'\n Description: {self.description or None}'
        if len(line) > 40: line = line[:40] + '...'
        info += line

        print(info)

    _ipython_display_ = show


    def register(self, print_msg=True):
        '''Add this impact indicator to the registry.'''
        self.registry.register_safely(self.ID, self)
        if print_msg:
            print(f'The impact indicator "{self.ID}" has been added to the registry.')


    def deregister(self, print_msg=True):
        '''Remove this impact indicator from the registry.'''
        self.registry.discard(self.ID)
        if print_msg:
            print(f'The impact indicator "{self.ID}" has been removed from the registry.')


    @classmethod
    def clear_registry(cls, print_msg=True):
        '''Remove all existing impact indicators from the registry.'''
        cls.registry.clear()
        if print_msg:
            print('All impact indicators have been removed from registry.')


    @classmethod
    def get_all_indicators(cls, include_alias=False):
        '''
        Get all defined impact indicator as a dict.

        Parameters
        ----------
        include_alias : bool
            If True, aliases will be included as keys in the dict as well.
        '''

        if not include_alias:
            return cls.registry.data

        else:
            dct = cls.registry.data.copy()
            dct.update(cls._get_alias_dct())
            return dct

    @classmethod
    def get_indicator(cls, ID_or_alias):
        '''Get an impact indicator by its ID or alias.'''
        dct = cls.get_all_indicators(True)
        return dct.get(ID_or_alias)

    @classmethod
    def load_indicators_from_file(cls, path_or_dict, index_col=None):
        '''Same as :func:`load_from_file`, has been deprecated.'''
        warn('`load_indicators_from_file` has been deprecated, '
             'please use `load_from_file` instead.', stacklevel=2)
        cls.load_from_excel(path_or_dict, index_col)

    @classmethod
    def load_from_file(cls, path_or_df, index_col=None):
        '''
        Load impact indicator from a datasheet.

        The first row of this datasheet should have "indicator"
        (it is used as the ID, e.g., GlobalWarming),
        "alias" (e.g., GWP), "unit" (e.g., kg CO2-eq), "method" (e.g., TRACI),
        "category" (e.g., environmental impact), and "description".
        Aside from "indicator", other information is optional.

        Each row should be a data entry.

        .. note::

            This function is just one way to batch-load impact indicators,
            you can always write your own function that fits your datasheet format,
            as long as it provides all the information to construct the impact indicator.


        Parameters
        ----------
        path_or_df : str or :class:`pandas.DataFrame`
            DataFrame or complete path of the datasheet, currently support tsv, csv, and xls/xlsx.
        index_col : None or int
            Index column of the :class:`pandas.DataFrame`.

        Tip
        ---
        [1] tsv is preferred as it shows up on GitHub.

        [2] Refer to the `Bwaise system <https://github.com/QSD-Group/EXPOsan/tree/main/exposan/bwaise/data>`_
        in the `Exposan` repository for a sample file.
        '''

        data = load_data(path=path_or_df, index_col=index_col) if isinstance(path_or_df, str) else path_or_df

        for num in data.index:
            new = cls.__new__(cls)
            kwargs = {}
            for k in ('alias', 'unit', 'method', 'category', 'description'):
                try:
                    kwargs[k] = data.iloc[num][k]
                except KeyError:
                    kwargs[k] = ''

            new.__init__(ID=data.iloc[num]['indicator'], **kwargs)

    @classmethod
    def _get_alias_dct(cls):
        dct = {}
        for i in cls.registry.data.values():
            if i.alias:
                dct[i.alias] = i
        return dct

    @property
    def ID(self):
        '''[str] ID of this impact indicator.'''
        return self._ID
    @ID.setter
    def ID(self, ID):
        self._ID = ID

    @property
    def alias(self):
        '''[str] Alias of this impact indicator.'''
        if not hasattr(self, '_alias'): # for initiation
            self._alias = None
        return self._alias
    @alias.setter
    def alias(self, alias):
        alias = None if str(alias) == 'nan' else alias
        alias_dct = self._get_alias_dct()

        if alias:
            if not isinstance(alias, str):
                raise TypeError(f'`alias` can only be a str, not {type(alias).__name__}.')

            if alias in alias_dct.keys():
                old_ind = alias_dct[alias]
                if old_ind.ID != self.ID:
                    warn(f'The alias "{alias}" is now being used for "{self.ID}", ' \
                         f'instead of {old_ind.ID}.')
                    old_ind._alias = None
            self._alias = alias

        else:
            self._alias = None

    @property
    def unit(self):
        '''[str] Unit of this impact indicator.'''
        return self._unit
    @unit.setter
    def unit(self, i):
        self._unit = str(i)
        self._ureg_unit, self._unit_remaining = parse_unit(i)

    @property
    def method(self):
        '''[str] Impact assessment method of this impact indicator.'''
        return self._method
    @method.setter
    def method(self, i):
        self._method = i

    @property
    def category(self):
        '''[str] Impact category of this impact indicator.'''
        return self._category
    @category.setter
    def category(self, i):
        self._category = i

    @property
    def description(self):
        '''[str] Description of this impact indicator.'''
        return self._description
    @description.setter
    def description(self, i):
        self._description = i

    @property
    def registered(self):
        '''[bool] If this impact indicator is registered in the record.'''
        data = self.registry.data.get(self.ID)
        return True if data else False


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''

import os
path = os.path.dirname(os.path.realpath(__file__))
data_path = path[:-6] + '/data/'
del os

import pandas as pd

__all__ = ('load_data', 'data_path', 'dct_from_str')


def load_data(path=None, sheet=None, index_col=0, **kwargs):
    '''For data importing.'''

    last_4 = path[-4:]
    if last_4 == '.tsv':
        data = pd.read_csv(path, sep='\t', index_col=index_col, **kwargs)
    elif last_4 == '.csv':
        data = pd.read_csv(path, index_col=index_col, **kwargs)
    elif last_4 in ('xlsx', '.xls'):
        try: data = pd.read_excel(path, sheet_name=sheet, engine='openpyxl',
                                  index_col=index_col, **kwargs)
        except: data = pd.read_excel(path, engine='openpyxl',
                                     index_col=index_col, **kwargs)
    else:
        raise ValueError('Only tsv, csv, or xlsx files can be loaded.')
    return data

def dct_from_str(dct_str, sep=',', dtype='float'):
    '''
    Use to parse str into a dict,
    the str should be written in `k1=v1, k2=v2, ..., kn=vn`
    (separated by comma or other symbols defined by `sep`).
    '''
    splitted = [i.split('=') for i in dct_str.replace(' ', '').split(sep)]

    if dtype == 'float':
        return {k:float(v) for k, v in splitted}

    elif dtype == 'int':
        return {k:int(v) for k, v in splitted}

    else:
        return {k:v for k, v in splitted}


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Smiti Mittal <smitimittal@gmail.com>
    Yalin Li <zoe.yalin.li@gmail.com>
    Anna Kogler

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''


from .. import Equipment

__all__ = ('Machine',)


class Machine(Equipment):
    '''
    Supplementary machines to be used in an electrochemical process.
    Refer to the example in :class:`ElectroChemCell` for how to use this class.

    Parameters
    ----------
    N : int
        Number of units of the given machine.
    unit_cost: float
        Unit cost of the machine

    See Also
    --------
    :class:`ElectroChemCell`

    '''
    __slots__ = ('_N', 'name', 'unit_cost')

    def __init__(self, name=None, # when left as None, will be the same as the class name
                 design_units={},
                 F_BM=1., lifetime=10000, lifetime_unit='hr', N=0,
                 unit_cost=0.1):
        Equipment.__init__(self=self, name=name, design_units=design_units,
                           F_BM=F_BM, lifetime=lifetime, lifetime_unit=lifetime_unit)
        self.name = name
        self.N = N
        self.unit_cost = unit_cost

    # All subclasses of `Machine` must have a `_design` and a `_cost` method
    def _design(self):
        design = {
            f'Number of {self.name}': self.N,
            }
        return design

    # All subclasses of `Membrane` must have a `_cost` method, which returns the
    # purchase cost of this equipment
    def _cost(self):
        return self.unit_cost*self.N


    @property
    def N(self):
        '''[int] Number of units of the electrode.'''
        return self._N
    @N.setter
    def N(self, i):
        self._N = int(i)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''


# %%

import qsdsan as qs
from datetime import date
from biosteam import TEA

__all__ = ('SimpleTEA',)

conflict_slots = ('lang_factor', 'system', 'units', 'feeds', 'products')

class SimpleTEA(TEA):
    '''
    Calculate an annualized cost for simple economic analysis that does not
    include loan payment (i.e., 100% equity) and taxes.

    Parameters
    ----------
    system : :class:`biosteam.System`
        The system this TEA is conducted for.
    discount_rate : float
        Interest rate used in discounted cash flow analysis.
    start_year : int
        Start year of the system.
    lifetime : int
        Total lifetime of the system, [yr]. Currently `biosteam` only supports int.

        .. note::

            As :class:`SimpleTEA` is a subclass of :class:`biosteam.TEA`,
            and :class:`biosteam.TEA` currently only supports certain
            depreciation schedules, lifetime must be larger than or equal to 6.


    uptime_ratio : float
        Fraction of time that the system is operating, should be in [0,1]
        (i.e., a system that is always operating has an uptime_ratio of 1).

        .. note::

            If a unit has an `uptime_ratio` that is different from the `uptime_ratio`
            of the system, the `uptime_ratio` of the unit will be used in calculating
            the additional operation expenses (provided in `unit.add_OPEX`).

            However, `uptime_ratio` of the unit will not affect the utility
            (heating, cooling, power) and material costs/environmental impacts.

            For example, if the `uptime_ratio` of the system and the unit are
            1 and 0.5, respectively, then in calculating operating expenses
            associated with the unit:

                - Utility and material costs/environmental impacts will be calculated for 1*24*365 hours per year.
                - Additional operaitng expenses will be calculated for 0.5*24*365 hours per year.

            If less utility and material flows are not used at the same `uptime_ratio`
            as the system, they should be normalized to be the same.
            For example, if the system operates 100% of time but a pump only works
            50% of the pump at 50 kW. Set the pump `power_utility` to be 50*50%=25 kW.


    CAPEX : float
        Capital expenditure, if not provided, is set to be the same as `installed_equipment_cost`.
    lang_factor : float or None
        A factor to estimate the total installation cost based on equipment purchase cost,
        leave as ``None`` if providing ``CAPEX``.
        If neither ``CAPEX`` nor ``lang_factor`` is provided,
        ``installed_equipment_cost`` will be calculated as the sum of purchase costs
        of all units within the system.
    annual_maintenance : float
        Annual maintenance cost as a fraction of fixed capital investment.
    annual_labor : float
        Annual labor cost.
    system_add_OPEX : float or dict
        Annual additional system-wise operating expenditure (on top of the `add_OPEX` of each unit).
        Float input will be automatically converted to a dict with the key being
        "System additional OPEX".
    construction_schedule : tuple or None
        Construction progress, must sum up to 1, leave as `None` will assume the system finishes within one year.

    Examples
    --------
    A system should be constructed prior to TEA, here we import a pre-constructed one.

    >>> import qsdsan as qs
    >>> from qsdsan.utils import load_example_cmps, load_example_sys
    >>> cmps = load_example_cmps()
    >>> sys = load_example_sys(cmps)
    >>> # Uncomment the line below to see the system diagram
    >>> # sys.diagram()
    >>> sys.simulate()
    >>> sys.show()
    System: sys
    ins...
    [0] water
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): H2O   111
                        NaCl  0.856
    [1] methanol
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Methanol  0.624
    [2] ethanol
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Ethanol  0.217
    outs...
    [0] alcohols
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Methanol  0.624
                        Ethanol   0.217
    [1] waste_brine
        phase: 'l', T: 350 K, P: 101325 Pa
        flow (kmol/hr): H2O   88.8
                        NaCl  0.684
    >>> tea = qs.SimpleTEA(system=sys, discount_rate=0.05, start_year=2021,
    ...                     lifetime=10, uptime_ratio=0.9,
    ...                     system_add_OPEX=0.03)
    >>> tea.show()
    SimpleTEA: sys
    NPV  : -258,730 USD at 5.0% discount rate
    EAC  : 33,507 USD/yr
    CAPEX: 60,912 USD (annualized to 7,888 USD/yr)
    AOC  : 25,618 USD/yr

    See Also
    --------
    `SanUnit and System <https://qsdsan.readthedocs.io/en/latest/tutorials/SanUnit_and_System.html>`_

    `TEA and LCA <https://qsdsan.readthedocs.io/en/latest/tutorials/TEA_and_LCA.html>`_
    '''

    __slots__ = (*(i for i in TEA.__slots__ if i not in conflict_slots),
                 '_system', '_units', '_feeds', '_products',
                 '_discount_rate', '_start_year', '_lifetime',
                 '_uptime_ratio', '_operating_hours', '_CAPEX', '_lang_factor',
                 '_annual_maintenance', '_annual_labor', '_system_add_OPEX')

    def __init__(self, system, discount_rate=0.05,
                 start_year=date.today().year, lifetime=10, uptime_ratio=1.,
                 CAPEX=0., lang_factor=None,
                 annual_maintenance=0., annual_labor=0., system_add_OPEX={},
                 construction_schedule=None):
        system.simulate()
        self.system = system
        system._TEA = self
        self.discount_rate = discount_rate
        # IRR (internal rate of return) is the discount rate when net present value is 0
        self.IRR = discount_rate
        self._IRR = discount_rate # guess IRR for solve_IRR method
        self._sales = 0 # guess cost for solve_price method
        self.start_year = start_year
        self.lifetime = lifetime
        self.uptime_ratio = 1.
        self._lang_factor = None
        self._CAPEX = CAPEX
        self.lang_factor = lang_factor
        self.annual_maintenance = annual_maintenance
        self.annual_labor = annual_labor
        self.system_add_OPEX = system_add_OPEX
        if not construction_schedule:
            construction_schedule = (1,)
        self.construction_schedule = construction_schedule

        ########## Not relevant to SimpleTEA but required by TEA ##########
        # From U.S. IRS for tax purpose, won't matter when tax set to 0
        # Based on IRS Publication 946 (2019), MACRS15 should be used for
        # municipal wastewater treatment plant, but the system lifetime is
        # just 10 yrs or shorter, so changed to a shorter one
        if lifetime >= 8:
            self.depreciation = 'MACRS7'
        elif lifetime >=6:
            self.depreciation = 'MACRS5'
        else:
             raise ValueError('Currently BioSTEAM does not support TEA ' \
                              'for systems with a lifetime shorter than 6 years.')

        self.income_tax = 0.
        self.startup_months = 0.
        self.startup_FOCfrac = 0.
        self.startup_VOCfrac = 0.
        self.startup_salesfrac = 0.
        self.WC_over_FCI = 0.
        self.finance_interest = 0.
        self.finance_years = 0
        self.finance_fraction = 0


    def __repr__(self):
        return f'<{type(self).__name__}: {self.system.ID}>'

    def show(self):
        c = self.currency
        info = f'{type(self).__name__}: {self.system.ID}'
        info += f'\nNPV  : {self.NPV:,.0f} {c} at {self.discount_rate:.1%} discount rate'
        info += f'\nEAC  : {self.EAC:,.0f} {c}/yr'
        info += f'\nCAPEX: {self.CAPEX:,.0f} {c} (annualized to {self.annualized_CAPEX:,.0f} {c}/yr)'
        info += f'\nAOC  : {self.AOC:,.0f} {c}/yr'
        print(info)

    _ipython_display_ = show


    def _DPI(self, installed_equipment_cost):
        return installed_equipment_cost

    def _TDC(self, DPI):
        return DPI

    def _FCI(self, TDC):
        return TDC

    def _FOC(self, FCI):
        return FCI*self.annual_maintenance+self.annual_labor+self.total_add_OPEX

    def get_unit_annualized_CAPEX(self, units):
        try: iter(units)
        except: units = (units, )
        CAPEX = 0
        r = self.discount_rate
        for unit in units:
            lifetime = unit.lifetime or self.lifetime
            # no unit lifetime or unit lifetime given as a number
            # (i.e., no individual equipment lifetime)
            if not isinstance(lifetime, dict):
                CAPEX += unit.installed_cost*r/(1-(1+r)**(-lifetime))
            else:
                lifetime_dct = dict.fromkeys(unit.purchase_costs.keys())
                lifetime_dct.update(lifetime)
                for equip, cost in unit.purchase_costs.items():
                    factor = unit.F_BM[equip]*\
                        unit.F_D.get(equip, 1.)*unit.F_P.get(equip, 1.)*unit.F_M.get(equip, 1.)
                    # for equipment that does not have individual lifetime
                    # use the unit lifetime or TEA lifetime
                    equip_lifetime = lifetime_dct[equip] or self.lifetime
                    CAPEX += factor*cost*r/(1-(1+r)**(-equip_lifetime))
        return CAPEX

    @property
    def system(self):
        '''[:class:`biosteam.System`] The system this TEA is conducted for.'''
        return self._system
    @system.setter
    def system(self, i):
        if i:
            self._system = i
            self._units = sorted([j for j in i.units if j._design or j._cost],
                                key=lambda x: x.line)
            self._feeds = i.feeds
            self._products = i.products

    @property
    def units(self):
        '''[:class:`qsdsan.SanUnit`] Units in the system.'''
        return self._units

    @property
    def feeds(self):
        '''[:class:`qsdsan.WasteStream`] System feed streams.'''
        return self._feeds

    @property
    def products(self):
        '''[:class:`qsdsan.WasteStream`] System product streams.'''
        return self._products

    @property
    def discount_rate(self):
        '''[float] Interest rate used in discounted cash flow analysis.'''
        return self._discount_rate
    @discount_rate.setter
    def discount_rate(self, i):
        if 0 <= i <= 1:
            self._discount_rate = float(i)
        else:
            raise ValueError('`discount_rate` must be in [0,1].')

    @property
    def start_year(self):
        '''[int] Start year of the system.'''
        return self._start_year
    @start_year.setter
    def start_year(self, i):
        self._start_year = i

    @property
    def lifetime(self):
        '''[int] Total lifetime of the system, [yr]. Currently `biosteam` only supports int.'''
        return int(self._lifetime)
    @lifetime.setter
    def lifetime(self, i):
        self._lifetime = self._years = int(i)
        self._duration = (int(self.start_year), int(self.start_year+self.lifetime))

    @property
    def duration(self):
        '''[int] Duration of the system based on start_year and lifetime.'''
        return self._duration

    @property
    def uptime_ratio(self):
        '''[float] Fraction of time that the system is operating.'''
        return self._uptime_ratio
    @uptime_ratio.setter
    def uptime_ratio(self, i):
        if 0 <=i<= 1:
            self._uptime_ratio = float(i)
            self._operating_days = 365*float(i)
            self._operating_hours = self._operating_days * 24
        else:
            raise ValueError('`uptime_ratio` must be in [0,1].')

    @property
    def operating_days(self):
        '''[float] Equivalent operating days calculated based on `uptime_ratio`.'''
        return self._operating_days

    @property
    def operating_hours(self):
        '''[float] Equivalent operating hours calculated based on `uptime_ratio`.'''
        return self._operating_hours

    @property
    def lang_factor(self):
        '''[float] A factor to estimate the total installation cost based on equipment purchase cost.'''
        return self._lang_factor or None
    @lang_factor.setter
    def lang_factor(self, i):
        if self.CAPEX is not None:
            if i is not None:
                raise AttributeError('`CAPEX` provided, `lang_factor` cannot be set. '
                                     'The calculated `lang_factor is` '
                                     f'{self.installed_equipment_cost/self.purchase_cost:.1f}.')
            else:
                self._lang_factor = None
        elif i >=1:
            self._lang_factor = float(i)
        else:
            raise ValueError('`lang_factor` must >= 1.')

    @property
    def currency(self):
        '''[str] TEA currency, same with `qsdsan.currency`.'''
        return qs.currency
    @currency.setter
    def currency(self, i):
        raise AttributeError('Currency can only be changed through `qsdsan.currency`.')

    @property
    def installed_equipment_cost(self):
        '''[float] Sum of installed cost of all units in the system, is the same as `CAPEX` if `CAPEX` is provided.'''
        if self._CAPEX:
            return self._CAPEX
        if self.lang_factor:
            return self.purchase_cost*self.lang_factor
        return sum([u.installed_cost for u in self.units])

    @property
    def DPI(self):
        '''[float] Direct permanent investment, same as `installed_equipment_cost`.'''
        return self._DPI(self.installed_equipment_cost)

    @property
    def TDC(self):
        '''[float] Total depreciable capital, same as `installed_equipment_cost`.'''
        return self._TDC(self.DPI)

    @property
    def FCI(self):
        '''[float] Fixed capital investment, same as `installed_equipment_cost`.'''
        return self._FCI(self.TDC)

    @property
    def TCI(self):
        '''[float] Total capital investment, same as `installed_equipment_cost`.'''
        return self.FCI

    @property
    def CAPEX(self):
        '''[float] Capital expenditure, if not provided, is set to be the same as `installed_equipment_cost`.'''
        return self.TCI

    @property
    def annual_maintenance(self):
        '''[float] Annual maintenance cost as a fraction of fixed capital investment.'''
        return self._annual_maintenance
    @annual_maintenance.setter
    def annual_maintenance(self, i):
        if 0 <= i <= 1:
            self._annual_maintenance = float(i)
        else:
            raise ValueError('`annual_maintenance` must be in [0,1].')

    @property
    def annual_labor(self):
        '''[float] Annual labor cost.'''
        return self._annual_labor
    @annual_labor.setter
    def annual_labor(self, i):
        self._annual_labor = float(i)

    @property
    def unit_add_OPEX(self):
        '''[float] Sum of `add_OPEX` for all units in the system.'''
        tot = 0
        for u in self.units:
            add_OPEX = sum(v for v in u.add_OPEX.values())
            tot += add_OPEX*u.uptime_ratio/self.uptime_ratio*self._operating_hours
        return tot

    @property
    def system_add_OPEX(self):
        '''
        [dict] Annual additional system-wise operating expenditure
        (on top of the `add_OPEX` of each unit).
        Float input will be automatically converted to a dict with the key being
        "System additional OPEX".
        '''
        return {'System dditional OPEX': self._system_add_OPEX} \
               if isinstance(self._system_add_OPEX, float) else self._system_add_OPEX
        return self._system_add_OPEX
    @system_add_OPEX.setter
    def system_add_OPEX(self, i):
        if isinstance(i, float):
            i = {'System additional OPEX': i}
        if not isinstance(i, dict):
            raise TypeError('system_add_OPEX can only be float of dict, ' \
                            f'not {type(i).__name__}.')
        self._system_add_OPEX = i

    @property
    def total_add_OPEX(self):
        '''[float] Sum of `unit_add_OPEX` and `system_add_OPEX`.'''
        return self.unit_add_OPEX+sum(v for v in self.system_add_OPEX.values())

    @property
    def FOC(self):
        '''
        [float] Fixed operating cost, including maintenance, labor, and any additional
        operaitng expenditure other than chemical inputs and utilities.
        '''
        return self._FOC(self.FCI)

    @property
    def annualized_CAPEX(self):
        '''[float] Annualized capital expenditure.'''
        return self.get_unit_annualized_CAPEX(self.units)


    @property
    def EAC(self):
        '''[float] Equivalent annual cost of the system (sales not included).'''
        return self.annualized_CAPEX+self.AOC


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Smiti Mittal <smitimittal@gmail.com>
    Yalin Li <zoe.yalin.li@gmail.com>
    Anna Kogler

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''


from .. import Equipment

__all__ = ('Electrode',)


class Electrode(Equipment):
    '''
    Electrodes to be used in an electrochemical cell.
    Refer to the example in :class:`ElectroChemCell` for how to use this class.

    Parameters
    ----------
    N : int
        Number of units of the given electrode.
    electrode_type : str
        Type of the electrode, can only be "anode" or "cathode".
    material: str
        Material of the electrode.
    unit_cost: float
        Unit cost of the electrode, will use default cost (if available)
        if not provided.
    surface_area : float
        Surface area of the electrode in m2.

    See Also
    --------
    :class:`ElectroChemCell`

    '''

    # Include all attributes (no properties) in addition to the ones in the
    # parent `Equipment` class
    # Using __slots__ can improve computational efficiency when the class does not
    # have many attributes
    __slots__ = ('_type', '_material', '_N', '_unit_cost', 'surface_area')

    _default_unit_cost = {'graphite': 50}


    def __init__(self, name=None, # when left as None, will be the same as the class name
                 design_units={}, # if no value, then should be an empty dict
                 F_BM=1.,
                 lifetime=10000, lifetime_unit='hr', N=0,
                 electrode_type='anode',
                 material='graphite', unit_cost=0.1, surface_area=1):
        Equipment.__init__(self=self, name=name, design_units=design_units, F_BM=F_BM, lifetime=lifetime, lifetime_unit=lifetime_unit)
        self.N = N
        self.electrode_type = electrode_type
        self.unit_cost = unit_cost
        self.material = material
        self.surface_area = surface_area


    # All subclasses of `Equipment` must have a `_design` and a `_cost` method
    def _design(self):
        design = {
            'Type of electrode' : self.electrode_type,
            f'Number of {self.electrode_type}': self.N,
            f'Material of {self.electrode_type}': self.material,
            f'Surface area of {self.electrode_type}': self.surface_area
            }
        self.design_units = {f'Surface area of {self.electrode_type}': 'm2'}
        return design

    # All subclasses of `Equipment` must have a `_cost` method, which returns the
    # purchase cost of this equipment
    def _cost(self):
        return self.unit_cost*self.N


    # You can use property to add checks
    @property
    def N(self):
        '''[str] Number of units of the electrode.'''
        return self._N
    @N.setter
    def N(self, i):
        self._N = int(i)

    @property
    def electrode_type(self):
        '''[str] Type of the electrode, either "anode" or "cathode".'''
        return self._type
    @electrode_type.setter
    def electrode_type(self, i):
        if i.lower() in ('anode', 'cathode'):
            self._type = i
        else:
            raise ValueError(f'Electrode can only be "anode" or "cathode", not "{i}".')

    @property
    def unit_cost(self):
        '''[float] Cost of one electrode.'''
        if self._unit_cost:
            return self._unit_cost
        cost = self._default_unit_cost.get(self.material)
        return cost or 0.
    @unit_cost.setter
    def unit_cost(self, i):
        self._unit_cost = i

    @property
    def material(self):
        '''[str] Material of the electrode.'''
        return self._material
    @material.setter
    def material(self, i):
        self._material = i.lower()


# -*- coding: utf-8 -*-
'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Joy Cheung <joycheung1994@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/master/LICENSE.txt
for license details.
'''

import os
import thermosteam as tmo
from thermosteam.utils import chemicals_user
# os.chdir("C:/Users/joy_c/Dropbox/PhD/Research/QSD/codes_developing/QSDsan")
from qsdsan import Components, Processes#, Process
from ..utils import data_path

__all__ = ('cmps_asm1', 'asm1', 'ASM1')


data_path = os.path.join(data_path, 'process_data/_asm1.tsv')
# data_path = "qsdsan/data/process_data/_asm1.tsv"

############# Components with default notation #############
cmps = Components.load_default()

S_I = cmps.S_U_Inf.copy('S_I')
S_I.description = 'Soluble inert organic matter'

#%%
X_I = cmps.X_U_Inf.copy('X_I')
X_I.description = 'Particulate inert organic matter'

S_S = cmps.S_F.copy('S_S')
S_S.description = 'Readily biodegradable substrate'
S_S.i_N = 0

X_S = cmps.X_B_Subst.copy('X_S')
X_S.description = 'Slowly biodegradable substrate'
X_S.i_N = 0

X_BH = cmps.X_OHO.copy('X_BH')
X_BH.description = 'Active heterotrophic biomass'

X_BA = cmps.X_AOO.copy('X_BA')
X_BA.description = 'Active autotrophic biomass'
X_BH.i_N = X_BA.i_N = 0.086     # i_XB

X_P = cmps.X_U_OHO_E.copy('X_P')
X_P.description = 'Particulate products arising from biomass decay'
X_P.i_N = 0.06                  # i_XP

S_O = cmps.S_O2.copy('S_O')

S_NO = cmps.S_NO3.copy('S_NO')
S_NO.description = 'Nitrate and nitrite nitrogen'

S_NH = cmps.S_NH4.copy('S_NH')

S_ND = cmps.S_F.copy('S_ND')
S_ND.description = 'Soluble biodegradable organic nitrogen'
S_ND.measured_as = 'N'
S_ND.i_COD = S_ND.i_C = S_ND.i_P = 0

X_ND = cmps.X_B_Subst.copy('X_ND')
X_ND.description = 'Particulate biodegradable organic nitrogen'
X_ND.measured_as = 'N'
X_ND.i_COD = X_ND.i_C = X_ND.i_P = 0

S_ALK = cmps.S_CO3.copy('S_ALK')      # measured as g C
S_ALK.description = 'Alkalinity, assumed to be HCO3-'

# add S_N2 to close mass balance
# add water for the creation of WasteStream objects
cmps_asm1 = Components([S_I, S_S, X_I, X_S, X_BH, X_BA, X_P,
                        S_O, S_NO, S_NH, S_ND, X_ND, S_ALK,
                        cmps.S_N2, cmps.H2O])

cmps_asm1.compile()
tmo.settings.set_thermo(cmps_asm1)

############ Processes in ASM1 #################
params = ('Y_H', 'Y_A', 'f_P',
          'mu_H', 'K_S', 'K_O_H', 'K_NO', 'b_H',
          'mu_A', 'K_NH', 'K_O_A', 'b_A',
          'eta_g', 'k_a', 'k_h', 'K_X', 'eta_h')

asm1 = Processes.load_from_file(data_path,
                                conserved_for=('COD', 'charge', 'N'),
                                parameters=params,
                                compile=True)

# ASM1 typical values at 20 degree C
asm1.set_parameters(
    Y_A = 0.24,                  # autotrophic yield = 0.24 gCOD/gN
    Y_H = 0.67,                  # heterotrophic yield = 0.67 gCOD/gCOD
    f_P = 0.08,                  # fraction of biomass yielding particulate products = 0.08, unitless
    mu_H = 6,                    # heterotrophic maximum specific growth rate = 6.0 d^(-1)
    K_S = 20,                    # readily biodegradable substrate half saturation coefficient = 20.0 gCOD/m3
    K_O_H = 0.2,                 # O2 half saturation coefficient = 0.2 gO2/m3
    K_NO = 0.5,                  # nitrate half saturation coefficient = 0.5 gN/m3
    b_H = 0.62,                  # heterotrophic biomass decay rate constant = 0.62 d^(-1)
    eta_g = 0.8,                 # reduction factor for anoxic growth of heterotrophs = 0.8, unitless
    eta_h = 0.4,                 # anoxic hydrolysis reduction factor = 0.4, unitless
    k_h = 3.0,                   # hydrolysis rate constant = 3.0 d^(-1)
    K_X = 0.03,                  # slowly biodegradable substrate half saturation coefficient for hydrolysis = 0.03 gCOD/gCOD
    mu_A = 0.8,                  # autotrophic maximum specific growth rate = 0.8 d^(-1)
    K_NH = 1.0,                  # ammonium (nutrient) half saturation coefficient = 1.0 gN/m3
    K_O_A = 0.4,                 # O2 half saturation coefficient for autotrophic growth = 0.4 gO2/m3
    b_A = 0.05,                  # !!! BSM1 value
    k_a = 0.08                   # ammonification rate constant = 0.08 d^(-1)/(gCOD/m3)
    )

# ASM1 typical values at 10 degree C
# asm1.set_parameters(
#     Y_A = 0.24,                  # autotrophic yield = 0.24 gCOD/gN
#     Y_H = 0.67,                  # heterotrophic yield = 0.67 gCOD/gCOD
#     f_P = 0.08,                  # fraction of biomass yielding particulate products = 0.08, unitless
#     mu_H = 3,                    # heterotrophic maximum specific growth rate = 3.0 d^(-1)
#     K_S = 20,                    # readily biodegradable substrate half saturation coefficient = 20.0 gCOD/m3
#     K_O_H = 0.2,                 # O2 half saturation coefficient = 0.2 gO2/m3
#     K_NO = 0.5,                  # nitrate half saturation coefficient = 0.5 gN/m3
#     b_H = 0.20,                  # heterotrophic biomass decay rate constant = 0.20 d^(-1)
#     eta_g = 0.8,                 # reduction factor for anoxic growth of heterotrophs = 0.8, unitless
#     eta_h = 0.4,                 # anoxic hydrolysis reduction factor = 0.4, unitless
#     k_h = 1.0,                   # hydrolysis rate constant = 1.0 d^(-1)
#     K_X = 0.01,                  # slowly biodegradable substrate half saturation coefficient for hydrolysis = 0.01 gCOD/gCOD
#     mu_A = 0.3,                  # autotrophic maximum specific growth rate = 0.3 d^(-1)
#     K_NH = 1.0,                  # ammonium (nutrient) half saturation coefficient = 1.0 gN/m3
#     K_O_A = 0.4,                 # O2 half saturation coefficient for autotrophic growth = 0.4 gO2/m3
#     b_A = 0.05,                  # !!! BSM1 value
#     k_a = 0.04                   # ammonification rate constant = 0.04 d^(-1)/(gCOD/m3)
#     )

@chemicals_user
class ASM1(Processes):
    '''
    Activated Sludge Model No. 1 in original notation.

    Parameters
    ----------
    components: class:`CompiledComponents`, optional
        Components corresponding to each entry in the stoichiometry array,
        defaults to thermosteam.settings.chemicals.
    Y_A : float, optional
        Autotrophic yield, in [g COD/g N]. The default is 0.24.
    Y_H : float, optional
        Heterotrophic yield, in [g COD/g COD]. The default is 0.67.
    f_P : float, optional
        Fraction of biomass yielding particulate products. The default is 0.08.
    i_XB : float, optional
        Nitrogen content of biomass, in [g N/g COD]. The default is 0.08.
    i_XP : float, optional
        Nitrogen content of particulate products arising from biomass decay,
        in [g N/g COD]. The default is 0.06.
    mu_H : float, optional
        Heterotrophic maximum specific growth rate, in [d^(-1)]. The default
        is 4.0.
    K_S : float, optional
        Readily biodegradable substrate half saturation coefficient, in [g COD/m^3].
        The default is 10.0.
    K_O_H : float, optional
        Oxygen half saturation coefficient for heterotrophic growth, in [g O2/m^3].
        The default is 0.2.
    K_NO : float, optional
        Nitrate half saturation coefficient, in [g N/m^3]. The default is 0.5.
    b_H : float, optional
        Heterotrophic biomass decay rate constant, in [d^(-1)]. The default is 0.3.
    eta_g : float, optional
        Reduction factor for anoxic growth of heterotrophs. The default is 0.8.
    eta_h : float, optional
        Reduction factor for anoxic hydrolysis. The default is 0.8.
    k_h : float, optional
        Hydrolysis rate constant, in [d^(-1)]. The default is 3.0.
    K_X : float, optional
        Slowly biodegradable substrate half saturation coefficient for hydrolysis,
        in [g COD/g COD]. The default is 0.1.
    mu_A : float, optional
        Autotrophic maximum specific growth rate, in [d^(-1)]. The default is 0.5.
    K_NH : float, optional
        Ammonium (nutrient) half saturation coefficient, in [g N/m^3]. The default
        is 1.0.
    b_A : float, optional
        Autotrophic biomass decay rate constant, in [d^(-1)]. The default is 0.05.
    K_O_A : float, optional
        Oxygen half saturation coefficient for autotrophic growth, in [g O2/m^3].
        The default is 0.4.
    k_a : float, optional
        Ammonification rate constant, in [m^3/g COD/d]. The default is 0.05.
    path : str, optional
        Alternative file path for the Gujer matrix. The default is None.

    References
    ----------
    .. [1] Henze, M.; Gujer, W.; Mino, T.; Loosdrecht, M. van. Activated Sludge
        Models: ASM1, ASM2, ASM2d and ASM3; IWA task group on mathematical modelling
        for design and operation of biological wastewater treatment, Ed.; IWA
        Publishing: London, 2000.
    .. [2] Rieger, L.; Gillot, S.; Langergraber, G.; Ohtsuki, T.; Shaw, A.; Takács,
        I.; Winkler, S. Guidelines for Using Activated Sludge Models; IWA Publishing:
        London, New York, 2012; Vol. 11.
        https://doi.org/10.2166/9781780401164.

    '''
    def __new__(cls, components=None, Y_A=0.24, Y_H=0.67, f_P=0.08, i_XB=0.08, i_XP=0.06,
                mu_H=4.0, K_S=10.0, K_O_H=0.2, K_NO=0.5, b_H=0.3, eta_g=0.8, eta_h=0.8,
                k_h=3.0, K_X=0.1, mu_A=0.5, K_NH=1.0, b_A=0.05, K_O_A=0.4, k_a=0.05,
                path=None, **kwargs):
        if not path: path = data_path
        self = Processes.load_from_file(path,
                                        conserved_for=('COD', 'charge', 'N'),
                                        parameters=params,
                                        components=components,
                                        compile=True)
        self._components.X_BH.i_N = self._components.X_BA.i_N = i_XB
        self._components.X_P.i_N = i_XP
        self.set_parameters(Y_A=Y_A, Y_H=Y_H, f_P=f_P, mu_H=mu_H, K_S=K_S, K_O_H=K_O_H,
                            K_NO=K_NO, b_H=b_H, eta_g=eta_g, eta_h=eta_h, k_h=k_h,
                            K_X=K_X, mu_A=mu_A, K_NH=K_NH, b_A=b_A, K_O_A=K_O_A, k_a=k_a,
                            **kwargs)
        return self


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Smiti Mittal <smitimittal@gmail.com>
    Yalin Li <zoe.yalin.li@gmail.com>
    Anna Kogler

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''

# %%

from .. import Equipment, SanUnit, WasteStream

__all__ = ('ElectrochemicalCell',)


class ElectrochemicalCell(SanUnit):
    '''
    Electrochemical cell for nutrient recovery.

    This unit has the following equipment:
        - :class:`~.equipments.Electrode`
        - :class:`~.equipments.Membrane`
        - :class:`~.equipments.Column`
        - :class:`~.equipments.Machine`

    Parameters
    ----------
    recovery : dict
        Keys refer to chemical component IDs. Values refer to recovery fractions (with 1 being 100%) for the respective chemicals.
    removal : dict
        Keys refer to chemical component IDs. Values refer to removal fractions (with 1 being 100%) for the respective chemicals.
    equipments : list
        List of Equipment objects part of the Electrochemical Cell.
    OPEX_over_CAPEX : float
        Ratio with which operating costs are calculated as a fraction of capital costs

    Example
    -------
    >>> # Set components
    >>> import qsdsan as qs
    >>> kwargs = dict(particle_size='Soluble',
    ...               degradability='Undegradable',
    ...               organic=False)
    >>> H2O = qs.Component.from_chemical('H2O', phase='l', **kwargs)
    >>> NH3 = qs.Component.from_chemical('NH3', phase='g', **kwargs)
    >>> NH3.particle_size = 'Dissolved gas'
    >>> NH4OH = qs.Component.from_chemical('NH4OH', phase='l', **kwargs)
    >>> H2SO4 = qs.Component.from_chemical('H2SO4', phase='l', **kwargs)
    >>> AmmoniumSulfate = qs.Component.from_chemical('AmmoniumSulfate', phase='l',
    ...                                              **kwargs)
    >>> CleaningAgent = qs.Component('CleaningAgent', MW=1, phase='l', **kwargs)
    >>> cmps = qs.Components((H2O, NH3, NH4OH, H2SO4, AmmoniumSulfate, CleaningAgent))
    >>> # Assuming all has the same molar volume as water for demonstration purpose
    >>> for cmp in cmps:
    ...     cmp.copy_models_from(H2O, names=['V'])
    ...     defaulted_properties = cmp.default()
    >>> qs.set_thermo(cmps)
    >>> # Set waste streams
    >>> influent = qs.WasteStream('influent', H2O=1000, NH4OH=50)
    >>> cleaning_agent = qs.WasteStream('cleaning_agent', price=5)
    >>> # Set equipments
    >>> anode = qs.equipments.Electrode(name='anode', N=1, electrode_type='anode',
    ...                                 material='graphite', surface_area=10)
    >>> cathode = qs.equipments.Electrode(name='cathode', N=1, electrode_type='cathode',
    ...                                   material='carbon', surface_area=10, unit_cost=1)
    >>> membrane = qs.equipments.Membrane(name='membrane', N=2,
    ...             material='polyethylene', unit_cost=0.2, surface_area=1)
    >>> column = qs.equipments.Column(name='column1', N=3,
    ...             material='resin', unit_cost=2, surface_area=20)
    >>> machine = qs.equipments.Machine(name='fan', N=1, unit_cost=3)
    >>> # Set the unit
    >>> U1 = qs.sanunits.ElectrochemicalCell('U1', ins=(influent, cleaning_agent),
    ...                                 outs=('rec', 'rem', 'leftover'),
    ...                                 recovery={'NH4OH':0.6}, removal={'NH4OH':0.2},
    ...                                 equipments=(anode, cathode, membrane, column, machine), OPEX_over_CAPEX = 0.2)
    >>> # Simulate and look at the results
    >>> U1.simulate()
    >>> # U1.diagram() # have a look at the diagram
    >>> U1.show() # doctest: +SKIP
    ElectrochemicalCell: U1
    ins...
    [0] influent
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (g/hr): H2O    1e+06
                     NH4OH  5e+04
        WasteStream-specific properties:
         pH         : 7.0
         TN         : 19424.5 mg/L
         TKN        : 19424.5 mg/L
    [1] cleaning_agent
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow: 0
        WasteStream-specific properties: None for empty waste streams
    outs...
    [0] rec
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (g/hr): NH4OH  3e+04
        WasteStream-specific properties:
         pH         : 7.0
         TN         : 775169.7 mg/L
         TKN        : 775169.7 mg/L
    [1] rem
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (g/hr): NH4OH  1e+04
        WasteStream-specific properties:
         pH         : 7.0
         TN         : 775169.7 mg/L
         TKN        : 775169.7 mg/L
    [2] leftover
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (g/hr): H2O    1e+06
                     NH4OH  1e+04
        WasteStream-specific properties:
         pH         : 7.0
         TN         : 3964.4 mg/L
         TKN        : 3964.4 mg/L
    >>> U1.results() # doctest: +SKIP
    Electrochemical cell                           Units            U1
    Design              Type of electrode                      cathode
                        Number of anode                              1
                        Material of anode                     graphite
                        Surface area of anode         m2            10
                        Number of cathode                            1
                        Material of cathode                     carbon
                        Surface area of cathode       m2            10
                        Number of membrane                           2
                        Material of membrane              polyethylene
                        Surface area of membrane      m2             1
                        Number of column1                            3
                        Material of column1                      resin
                        Surface area of column1       m2            20
                        Number of fan                                1
    Purchase cost       anode                        USD           0.1
                        cathode                      USD             1
                        membrane                     USD           0.4
                        column1                      USD           120
                        fan                          USD             3
    Total purchase cost                              USD           124
    Utility cost                                  USD/hr             0
    Additional OPEX                               USD/hr          24.9
    '''

    def __init__(self, ID='', ins=None, outs=(), recovery={'NH3':0.6}, removal={'NH3':0.2},
                 equipments=(), OPEX_over_CAPEX=0):
        if isinstance(equipments, Equipment):
            equipments = (equipments,)
        SanUnit.__init__(self=self, ID=ID, ins=ins, outs=outs, equipments=equipments)
        self.recovery = recovery
        self.removal = removal
        self.OPEX_over_CAPEX = OPEX_over_CAPEX

    _N_ins = 2
    _N_outs = 3

    def _run(self):
        influent, cleaner = self.ins
        recovered, removed, left = self.outs[0], self.outs[1], self.outs[2]

        mixture = WasteStream()
        mixture.mix_from(self.ins)
        left.copy_like(mixture)

        for chemical, ratio in self.recovery.items():
            recovered.imass[chemical] = mixture.imass[chemical]*ratio
            left.imass[chemical] = left.imass[chemical]-mixture.imass[chemical]*ratio

        for chemical, ratio in self.removal.items():
            removed.imass[chemical] = mixture.imass[chemical]*ratio
            left.imass[chemical] = left.imass[chemical]-mixture.imass[chemical]*ratio

    def _design(self):
        self.add_equipment_design()

    def _cost(self):
        self.add_equipment_cost()
        self.equip_costs = self.baseline_purchase_costs.values()
        add_OPEX = sum(self.equip_costs)*self.OPEX_over_CAPEX
        self._add_OPEX = {'Additional OPEX': add_OPEX}


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''


# %%

from collections.abc import Iterable
from biosteam._graphics import UnitGraphics
from .. import SanUnit

__all__ = ('LumpedCost',)

class LumpedCost(SanUnit):
    '''
    A unit that does not affect stream flow (e.g., outs will be copied from ins),
    but only for cost calculation purpose.

    Parameters
    ----------
    cost_item_name : str
        The name of this unit that will be shown in design, cost, and result dicts.
    CAPEX : float
        Total installed capital cost.
    power : float
        Total electricity usage.
    add_OPEX : float
        Additional operating cost per hour.

    Examples
    --------
    `bwaise systems <https://github.com/QSD-Group/EXPOsan/blob/main/exposan/bwaise/systems.py>`_
    '''

    def __init__(self, ID='', ins=None, outs=(), thermo=None, init_with='WasteStream',
                 cost_item_name='Lumped cost',
                 CAPEX=0., power=0., add_OPEX=0., **kwargs):
        if isinstance(ins, str) or (not isinstance(ins, Iterable)):
            self._N_outs = self._N_ins = 1
        else:
            self._N_outs = self._N_ins = len(ins)
        self._graphics = UnitGraphics.box(self._N_ins, self._N_outs)

        SanUnit.__init__(self, ID, ins, outs, thermo, init_with)
        self.F_BM = {cost_item_name: 1}
        self.baseline_purchase_costs = {cost_item_name: CAPEX}
        self.power_utility(power)
        self._add_OPEX = add_OPEX
        for attr, val in kwargs.items():
            setattr(self, attr, val)

    def _run(self):
        for num, stream in enumerate(self.ins):
            self.outs[num].copy_like(stream)

    def _setup(self):
        pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''


# %%

from biosteam.utils import NotImplementedMethod
from .utils import auom

__all__ = ('Equipment',)

hasattr = hasattr

class Equipment:
    '''
    A flexible class for the design of individual equipment of a :class:`SanUnit`,
    this class can be dependent on but will not affect the mass flows within
    the unit.

    A non-abstract instance of this class must have:

        - A :func:`Equipment._design` method for equipment design.

            - This method should be called in the `_design` method of the unit
              the equipment belongs to using :func:`SanUnit.add_equipment_design`.
            - It should only take the unit this equipment belongs to as the parameter.
            - It should return a dict that contains the design (e.g., dimensions)
              of this equipment.
            - Unit (e.g., m, kg) of the design parameters should be stored in
              the attribute `Equipment._units`.

        - A :func:`Equipment._cost` method for equipment cost.

            - This method should be called in the `_cost` method of the unit
              the equipment belongs to using :func:`SanUnit.add_equipment_cost`.
            - It should only take the unit this equipment belongs to as the parameter.
            - It should return a float that contains the total purchase cost of this
              equipment.
            - Installed cost (:math:`C_{BM}`) of this equipment will be caluculated \
            based on the purchase cost (:math:`C_{Pb}`)

                .. math::

                   C_{BM} = C_{Pb} (F_{BM} + F_{D}F_{P}F_{M} - 1)


    Parameters
    ----------
    Name : str
        Name of this equipment, can be left as None.
    design_units: dict
        Unit (e.g., m, kg) of design parameters.
    F_BM: float
        Bare module factor of this equipment.
    F_D: float
        Design factor of this equipment.
    F_P: float
        Pressure factor of this equipment.
    F_M: float
        Material factor of this equipment.
    lifetime: float
        Lifetime of this equipment.
    lifetime_unit: str
        Unit of the lifetime.

    '''

    __slots__ = ('_linked_unit', 'name', 'design_units', 'F_BM', 'F_D', 'F_P', 'F_M',
                 'lifetime')

    def __init_subclass__(self, isabstract=False):
        if isabstract: return
        for method in ('_design', '_cost'):
            if not hasattr(self, method):
                raise NotImplementedError(
                    f'Equipment subclasses must have a {method} method unless `isabstract` is True.')

    _design = _cost = NotImplementedMethod

    def __init__(self, name=None, design_units=dict(), F_BM=1., F_D=1., F_P=1., F_M=1.,
                 lifetime=None, lifetime_unit='yr', **kwargs):

        if 'BM' in kwargs.keys():
            raise DeprecationWarning('`BM` has been depreciated, please use ' \
                                     f'`F_BM` for the Equipment {name}.')

        self._linked_unit = None
        self.name = name
        self.design_units = design_units
        self.F_BM = F_BM
        self.F_D = F_D
        self.F_P = F_P
        self.F_M = F_M
        self.lifetime = auom(lifetime_unit).convert(lifetime, 'yr')


    def __repr__(self):
        line = f'<{type(self).__name__}'
        line += f': {self.name}' if self.name else ''
        line += f' in {self.linked_unit}>' if self.linked_unit else '>'
        return line

    @property
    def linked_unit(self):
        '''
        :class:`SanUnit` The unit that this equipment belongs to.

        .. note::

            This property will be updated upon initiation of the unit.
        '''
        return self._linked_unit

    @property
    def design(self):
        '''[dict] Design information generated by :func:`Equipment._design`.'''
        return self._design()

    @property
    def purchase_cost(self):
        '''[float] Total purchase cost generated by :func:`Equipment._cost`.'''
        return self._cost()

    @property
    def installed_cost(self):
        '''[float] Total installed cost based on purchase cost and bare module factor.'''
        return self.purchase_cost*(self.F_BM+self.F_D*self.F_P*self.F_M-1)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''

# This tiered importing is because some modules in utils need to be imported
# before the the main modules (e.g., _component) since the main modules depend
# on them, while other modules in utils depend on the main modules.

from . import (
    cod,
    decorators,
    # descriptors, # currently not in use
    getters,
    formatting,
    loading,
    parse,
    setters,
    units_of_measure,
    )

from .cod import *
from .decorators import *
# from .descriptors import *
from .getters import *
from .formatting import *
from .loading import *
from .parse import *
from .setters import *
from .units_of_measure import *


__all__ = (
    *cod.__all__,
    *decorators.__all__,
    # *descriptors.__all__,
    *getters.__all__,
    *formatting.__all__,
    *loading.__all__,
    *parse.__all__,
    *setters.__all__,
    *units_of_measure.__all__,
    )


def _secondary_importing():
    from . import (
        doc_examples,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''


# from collections import defaultdict
from weakref import WeakKeyDictionary

__all__ = ('Validator', 'NonNegativeFloat', 'NonNgeativeInt', 'Fraction',
           # 'BareModule'
           )


# %%

class Validator:
    '''
    Descriptors can be used to make resusable property logics,
    a great reference can be found online. [1]_
    
    .. note::
        
        Using Descriptors to check value properties is slightly slower than using
        the decorators functions in qsdsan.utils.checkers.
    
    
    References
    ----------
    .. [1] https://nbviewer.jupyter.org/urls/gist.github.com/ChrisBeaumont/5758381/raw/descriptor_writeup.ipynb
     
    '''
    
    __slots__ = ['default', 'name', 'instance', 'owner']
    
    def __init__(self, default=None, name=None):
        self.default = default
        self.name = name
        self.data = WeakKeyDictionary()
        
    def __get__(self, instance, owner):
        return self.data.get(instance, self.default)


class NonNegativeFloat(Validator):
    '''For non-negative floats.'''
    
    def __set__(self, instance, value):
        if value < 0:
            if self.name:
                raise ValueError(f'Value for {self.name} cannot be negative.')
            else:
                raise ValueError('Value cannot be negative.')
        self.data[instance] = float(value)


class NonNgeativeInt(Validator):
    '''For non-negative integers.'''
    
    def __set__(self, instance, value):
        if value < 0:
            if self.name:
                raise ValueError(f'Value for {self.name} cannot be negative.')
            else:
                raise ValueError('Value cannot be negative.')
        self.data[instance] = int(value)


class Fraction(Validator):
    '''For values in [0,1].'''
    
    def __set__(self, instance, value):
        if not 0 <= value <= 1:
            if self.name:
                raise ValueError(f'Value for {self.name} must be in [0,1].')
            else:
                raise ValueError('Value must be in [0,1].')
        self.data[instance] = float(value)
        


# # Not in use
# class BareModule:
#     '''For unit BM (bare module factors).'''
    
#     def __get__(self, obj, objtype=None):
#         # Copy the _BM dict from the class attribute to prevent overwriting,
#         # need to in both __get__ and __set__ as not sure which one will be called first
#         if '_BM' not in obj.__dict__:
#             _BM = defaultdict(lambda:1)
#             _BM.update(obj._BM)
#             obj._BM = _BM
        
#         # If don't want defaults
#         # if '_BM' not in obj.__dict__:
#         #     obj._BM = obj._BM.copy()

#         return obj._BM
    
#     def __set__(self, obj, BM:dict):
#         # Copy the _BM dict from the class attribute to prevent overwriting,
#         # need to in both __get__ and __set__ as not sure which one will be called first
#         if '_BM' not in obj.__dict__:
#             _BM = defaultdict(lambda:1)
#             _BM.update(obj._BM)
#             obj._BM = _BM
        
#         # If don't want defaults
#         # if '_BM' not in obj.__dict__:
#         #     obj._BM = obj._BM.copy()
        
#         # if not isinstance(BM, dict):
#         #     raise TypeError('`BM` must be a dict, not {type(i).__name__}.')
        
#         # for i in self.BM.keys():
#         #     if i not in self.purhcase_costs.keys():
#         #         raise ValueError(f'The item "{i}" does not exist in `purchase_costs`.')

#         obj._BM.update(BM)



        
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Smiti Mittal <smitimittal@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''

from ._electrode import *
from ._membrane import *
from ._column import *
from ._machine import*


from . import (
    _electrode,
    _membrane,
    _column,
    _machine
    )


__all__ = (
    *_electrode.__all__,
    *_membrane.__all__,
    *_column.__all__,
    *_machine.__all__,
           )